# -*- coding: utf-8 -*-
"""
Contains the :py:class:`~radis.lbl.factory.SpectrumFactory` class, which is 
the core of the RADIS Line-by-Line module. 

Examples
--------

Calculate a CO Spectrum, fetching the lines from HITRAN ::

    # This is how you get a spectrum (see spectrum.py for front-end functions
    # that do just that)
    sf = SpectrumFactory(2125, 2249.9,
                         parallel=False,bplot=False,
                         molecule='CO', 
                         isotope=1,
                         cutoff=1e-30,   # for faster calculations. See
                                         # `plot_linestrength_hist` for more details
                         **kwargs)
    sf.fetch_databank()        # autodownload from HITRAN 
    s = sf.eq_spectrum(Tgas=300)
    s.plot('abscoeff')

    # Here we get some extra informations:
    s.plot('radiance', wunit='nm',
                       Iunit='µW/cm2/sr/nm',  # Iunit is arbitrary. Use whatever makes sense
                       show_points=True)  # show_points to have an idea of the resolution

See :py:mod:`radis.test.lbl.test_factory` for more examples


Routine Listing
---------------
*(the following sections are marked on the slider bar in the Spyder IDE with
the # XXX symbol)*

PUBLIC METHODS

- :meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`             >>> calc equilibrium spectrum
- :meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`         >>> calc non equilibrium spectrum
- :meth:`~radis.lbl.factory.SpectrumFactory.optically_thin_power`    >>> get total power (equilibrium or non eq)

Most methods are written in inherited class with the following inheritance scheme:
    
:py:class:`~radis.lbl.loader.DatabankLoader` > :py:class:`~radis.lbl.base.BaseFactory` > 
:py:class:`~radis.lbl.broadening.BroadenFactory` > :py:class:`~radis.lbl.bands.BandFactory` > 
:py:class:`~radis.lbl.factory.SpectrumFactory` > :py:class:`~radis.lbl.parallel.ParallelFactory`

Notes
-----

Debugging:

Remember one can access all locals variable in a function either by running
%debug after a crash in IPython. Or by adding the line `globals().update(locals())`
at any point


Performance:

Fast version: iterate over chunks of dataframe
Note that we can't use the full dataframe because it's too big and takes too much memory
See :ref:`Performance <label_lbl_performance>` for more details. 


for Developers:

- To implement new database formats, see the databases parsers in cdsd.py / hitran.py,
  and the partition function interpolators / calculators methods of SpectrumFactory:
  :py:meth:`~radis.lbl.loader.DatabankLoader._build_partition_function_calculator` and 
  :py:meth:`~radis.lbl.loader.DatabankLoader._build_partition_function_interpolator`


----------

"""
from __future__ import print_function, absolute_import, division, unicode_literals
from six import string_types
from warnings import warn
from radis.io.hitran import get_molecule
from radis.lbl.bands import BandFactory
from radis.lbl.base import get_all_waveranges
from radis.spectrum.spectrum import Spectrum
from radis.spectrum.equations import calc_radiance
from radis.misc.basics import is_float, list_if_float
from radis.phys.convert import conv2
from radis.phys.constants import k_b
from radis.phys.units import convert_rad2nm, convert_emi2nm
from radis import get_version
from numpy import exp, arange
from multiprocessing import cpu_count
from time import time
import numpy as np

# %% Main functions


class SpectrumFactory(BandFactory):
    ''' A class to put together all functions related to loading CDSD / HITRAN
    databases, calculating the broadenings, and summing over all the lines

    Parameters
    ----------

    wavenum_min: cm-1
        minimum wavenumber to be processed in cm^-1. Note that this wavenumber
        can be in 'air' or 'vacuum' depending on the value of the parameter
        "medium="

    wavenum_max: cm-1
        maximum wavenumber to be processed in cm^-1

    wavelength_min: nm
        minimum wavelength to be processed in nm

    wavelength_max: nm
        maximum wavelength to be processed in nm

    Tref: K
        reference temperature for calculations, HITRAN database uses 296 Kelvin
        default: 3400K

    pressure: bar
        partial pressure of gas in bar. Default 1.01325 (1 atm)

    mole_fraction: N/D
        species mole fraction. Default 1. Note that the rest of the gas
        is considered to be air for collisional broadening.

    path_length: cm
        path length in cm. Default 1.

    molecule: int, str, or ``None``
        molecule id (HITRAN format) or name. If None, the molecule is infered
        from the database files being loaded. It is just good practice to fill
        it. Default ``None``.

    isotope: int, list, str of the form '1,2', or 'all'
        isotope id (sorted by relative density: (eg: 1: CO2-626, 2: CO2-636 for CO2).
        See HITRAN documentation for isotope list for all species. If 'all',
        all isotopes in database are used (this may result in larger computation
        times!). Default 'all'

    medium: ``'air'``, ``'vacuum'``
        propagating medium: choose whether to return wavelength in air or vacuum.
        Note that the  input wavelength range ("wavenum_min", "wavenum_max") changes.
        Default ``'air'``

    Other Parameters
    ----------------

    Computation parameters:

    Tref: K
        Reference temperature for calculations (linestrength temperature
        correction). HITRAN database uses 296 Kelvin. Default 296 K

    self_absorption: boolean
        self absorption

    broadening_max_width: cm-1
        Full width over which to compute the broadening. Large values will create
        a huge performance drop (because convolutions are not vectorized).
        Also, the calculated spectral range is increased (by broadening_max_width/2
        on each side) to take into account overlaps from out-of-range lines.
        Default ``10`` cm-1.

    wstep: cm-1
        Spacing of calculated spectrum. Default 0.01 cm-1

    cutoff: float (~ unit of Linestrength: cm-1/(#.cm-2))
        discard linestrengths that are lower that this, to reduce calculation
        times. 1e-27 is what is generally used to generate databases such as
        CDSD. If 0, no cutoff. Default 1e-27.

    pseudo_continuum_threshold: float
        if not 0, first calculate a rough approximation of the spectrum, then
        moves all lines whose linestrength intensity is less than this threshold
        of the maximum in a semi-continuum. Values above 0.01 can yield significant
        errors, mostly in highly populated areas. 80% of the lines can typically
        be moved in a continuum, resulting in 5 times faster spectra. If 0,
        no semi-continuum is used. Default 0.

    bplot: boolean
        plot intermediary results (like slit function generation). Default ``False``.

    save_memory: boolean
        if ``True``, removes databases calculated by intermediate functions (for
        instance, delete the full database once the linestrength cutoff criteria
        was applied). This saves some memory but requires to reload the database
        & recalculate the linestrength for each new parameter. Default ``False``.

    export_populations: ``'vib'``, ``'rovib'``, ``None``
        if not None, store populations in Spectrum. Either store vibrational
        populations ('vib') or rovibrational populations ('rovib'). Default ``None``

    export_lines: boolean
        if ``True``, saves lines in Spectrum. Default ``True``.

    parallel: boolean
        use parallel compution. May only offer a performance improvement if
        one spectrum already takes > 10s to be calculated. Default ``False`` because
        since the latest vectorized version parallelisation is mostly
        counterproductive.

    db_use_cached: boolean, or 'regen'
        use cache for line databases.
        If ``True``, a pandas-readable csv file is generated on first access,
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot. But! ... be sure to delete these files
        to regenerate them if you happen to change the database. If 'regen',
        existing cached files are removed and regenerated. Default ``False``
        From 0.9.16 it is also used to load energy levels from .h5 cache file
        if exist

    lvl_use_cached: boolean, or ``'regen'``, or ``'force'``, or ``None``
        use cache for energy level database
        If None, the same value as ``db_use_cached`` is used.

    Nprocs, Ngroups: int
        parameters used in parallel processing mode. Default ``None``

    chunksize: float, or ``None``
        Splits the lines database in several chuncks during calculation, else
        the multiplication of lines over all spectral range takes too much memory
        and slows the system down. Chunksize let you change the default chunck
        size. If ``None``, all lines are processed directly. Usually faster but
        can create memory problems. Default ``None``


    Examples
    --------
    
    An example using :class:`~radis.lbl.factory.SpectrumFactory`, 
    :meth:`~radis.lbl.loader.DatabankLoader.load_databank`, and the 
    :class:`~radis.spectrum.spectrum.Spectrum` methods::

        sf = SpectrumFactory(wavelength_min=4150, wavelength_max=4400,
                             parallel=False, bplot=False, cutoff=1e-25,
                             isotope='1,2', db_use_cached=True,
                             broadening_max_width=10,
                             pseudo_continuum_threshold=0,
                             medium='vacuum',
                             verbose=verbose)
        sf.load_databank('CDSD-HITEMP')        # predefined in ~/.radis
        s = sf.eq_spectrum(Tgas=300, path_length=1)
        s.rescale_path_length(0.01)
        s.plot('radiance_noslit', Iunit='µW/cm2/sr/nm')

    Refer to the online :ref:`Examples <label_examples>` for more cases. 

    See Also
    --------

    :func:`~radis.lbl.calc.calc_spectrum`, 
    :class:`~radis.lbl.parallel.ParallelFactory`

    Main Methods: 
        
    :meth:`~radis.lbl.loader.DatabankLoader.load_databank`,
    :meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`,
    :meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`
    
    For advanced use:
        
    :meth:`~radis.lbl.loader.DatabankLoader.fetch_databank`,
    :meth:`~radis.lbl.loader.DatabankLoader.init_databank`,
    :meth:`~radis.lbl.loader.DatabankLoader.init_database`,
    :meth:`~radis.lbl.band.BandFactory.eq_bands`, 
    :meth:`~radis.lbl.band.BandFactory.non_eq_bands`
    
    
    '''
    # TODO: make it possible to export both 'vib' and 'rovib'


    # TODO
    # -------
    #
    # This code is still designed for one-molecule only: some parameters
    # (e.g.: molar mass used in Doppler broadening) are defined in SpectrumFactory
    # rather than being a column in the dataframe. This should be changed to
    # allow dealing with polymolecules databases

    # TODO
    # store everything in a self.var class instead of self.[] directly

    def __init__(self, wavenum_min=None,
                 wavenum_max=None,
                 wavelength_min=None,
                 wavelength_max=None,
                 Tref=296,
                 pressure=1.01325,
                 mole_fraction=1,
                 path_length=1,
                 wstep=0.01,
                 molecule=None,
                 isotope='all',
                 medium='air',
                 broadening_max_width=10,
                 pseudo_continuum_threshold=0,
                 self_absorption=True,
                 chunksize=None,
                 Nprocs=None,
                 Ngroups=None,
                 cutoff=1e-27,
                 bplot=False,
                 parallel=False,
                 db_use_cached=True,
                 lvl_use_cached=None,
                 verbose=True,
                 warnings=True,
                 save_memory=False,
                 export_populations='vib',
                 export_lines=True,
                 **kwargs):

        # Initialize BandFactory object
        super(SpectrumFactory, self).__init__()

        # Check inputs (deal with deprecated format)
        if medium not in ['air', 'vacuum']:
            raise ValueError("Wavelength must be one of: 'air', 'vacuum'")
        kwargs0 = kwargs     # kwargs is used to deal with Deprecated names
        if 'use_cached' in kwargs:
            warn(DeprecationWarning('use_cached replaced with db_use_cached'))
            db_use_cached = kwargs0.pop('use_cached')
        if 'isotope_identifier' in kwargs:
            warn(DeprecationWarning('isotope_identifier replaced with isotope'))
            isotope = kwargs0.pop('isotope_identifier')
        if 'air_pressure_mbar' in kwargs:
            raise DeprecationWarning(
                'air_pressure_mbar replaced with pressure (in bar!)')
        if 'slit_function_FWHM' in kwargs:
            raise DeprecationWarning('slit_function_FWHM removed from inputs. Use ' +
                                     'Spectrum.apply_slit() directly')
        if kwargs0 != {}:
            raise TypeError("__init__() got an unexpected keyword argument '{0}'".format
                            (list(kwargs0)[0]))

        if not 0 <= pseudo_continuum_threshold < 1:
            raise ValueError('pseudo_continuum_threshold should be in [0-1]')
        if (export_populations not in ['vib', 'rovib', False, None] and 
            not isinstance(export_populations, list)):
            raise ValueError("export_populations must be one of 'vib', 'rovib', "+\
                             "or 'False'. Got '{0}'".format(
                    export_populations))

        # calculate waveranges
        # --------------------

        # Get wavelength & wavenumbers in correct propagation medium
        wavenum_min, wavenum_max, wavelength_min, wavelength_max = get_all_waveranges(
            medium, wavenum_min, wavenum_max, wavelength_min, wavelength_max)

        # calculated range is broader than output waverange to take into account off-range line broadening
        wavenumber, wavenumber_calc = _generate_wavenumber_range(wavenum_min, wavenum_max,
                                                                 wstep, broadening_max_width)
        wbroad_centered = _generate_broadening_range(
            wstep, broadening_max_width)

        # Get boolean array to switch from `wavenumber_calc` to `wavenumber`
        woutrange = np.in1d(wavenumber_calc, wavenumber, assume_unique=True)
        self.wbroad_centered = wbroad_centered
        self.wavenumber = wavenumber
        self.wavenumber_calc = wavenumber_calc
        self.woutrange = woutrange

        # Init variables
        # --------------

        # Get molecule name
        if isinstance(molecule, int):
            molecule == get_molecule(molecule)

        # Store isotope identifier in str format (list wont work in database queries)
        if not isinstance(isotope, string_types):
            isotope = ','.join([str(k) for k in list_if_float(isotope)])

        # Initialize input conditions
        self.input.wavenum_min = wavenum_min
        self.input.wavenum_max = wavenum_max
        self.input.wavelength_min = wavelength_min
        self.input.wavelength_max = wavelength_max
        # propagating medium. Calculations are done in vacuum anyway
        self.input.medium = medium
        self.input.Tref = Tref
        self.input.pressure_mbar = pressure*1e3
        self.input.mole_fraction = mole_fraction

        self.input.path_length = path_length
        self.input.molecule = molecule  # if None, will be overwritten after reading database
        self.input.state = 'X'              # for the moment only ground-state is used
        # (but the code is electronic state aware)
        self.input.isotope = isotope    # if 'all', will be overwritten after reading database
        self.input.self_absorption = self_absorption

        # Initialize computation variables
        self.params.wstep = wstep
        self.params.db_use_cached = db_use_cached
        if lvl_use_cached is None:
            lvl_use_cached = db_use_cached
        self.params.lvl_use_cached = lvl_use_cached
        self.params.pseudo_continuum_threshold = pseudo_continuum_threshold
        self.misc.parallel = parallel
        self.params.cutoff = cutoff
        self.params.broadening_max_width = broadening_max_width   # line broadening
        self.misc.export_lines = export_lines
        self.misc.export_populations = export_populations
        self.params.wavenum_min_calc = wavenumber_calc[0]
        self.params.wavenum_max_calc = wavenumber_calc[-1]

        # used to split lines into blocks not too big for memory
        self.misc.chunksize = chunksize
        if parallel:
            # Set up parameters for //
            if Ngroups is None:
                Ngroups = cpu_count()  # number of processors
            if Nprocs is None:
                Nprocs = cpu_count()
            self.misc.Ngroups = Ngroups
            self.misc.Nprocs = Nprocs
        else:
            if Nprocs is not None:
                print('Choose parallel=True to use Nprocs')
            if Ngroups is not None:
                print('Choose parallel=True to use Ngroups')

        # Other parameters:
        self.bplot = bplot
        self.verbose = verbose
        self.save_memory = save_memory
        self.autoupdatedatabase = False  # a boolean to automatically store calculated
        # spectra in a Spectrum database. See init_database
        # for more info
        self.autoretrievedatabase = False  # a boolean to automatically retrieve
        # spectra from database instead of
        # calculating them
        self.SpecDatabase = None   # the database to store spectra. Not to be confused
        # with the databank where lines are stored
        self.database = None       # path to previous database

        # Warnings
        # --------

        # Add cutoff / threshold warning values for the different optimisations
        # (linestrength cutoff, broadening cutoff). These cannot be changed
        # from the Factory input, but can still be modified manually afterwards
        # TODO: replace everything with 'auto' modes.

        # Set default behavior for warnings:
        if warnings in [True, 'warning']:
            self.warnings['default'] = 'warning'
        elif warnings == 'error':
            self.warnings['default'] = 'error'
        elif warnings in [False, 'ignore']:
            self.warnings['default'] = 'ignore'
        else:
            raise ValueError(
                'Unexpected value for warnings: {0}'.format(warnings))
        # Set default values for warnings thresholds
        self.misc.warning_linestrength_cutoff = 1e-2
        self.misc.warning_broadening_threshold = 1e-2

        return

    # %% ======================================================================
    # PUBLIC METHODS
    # ------------------------
    # eq_spectrum             >>> calc equilibrium spectrum
    # non_eq_spectrum         >>> calc non equilibrium spectrum
    # eq_bands                >>> returns bands as a list of spectra
    # non_eq_bands            >>> same, with overpopulation factors
    # power                   >>> get total power (equilibrium or non eq)
    #
    # XXX =====================================================================

    def eq_spectrum(self, Tgas, mole_fraction=None, path_length=None,
                    pressure=None, name=None):
        ''' Generate a spectrum at equilibrium

        Parameters
        ----------

        Tgas: float
            Gas temperature (K)

        mole_fraction: float
            database species mole fraction. If None, Factory mole fraction is used.

        path_length: float
            slab size (cm). If None, Factory mole fraction is used.

        pressure: float
            pressure (bar). If None, the default Factory pressure is used.

        name: str
            case name (useful in batch)

        Returns
        -------

        Returns a :class:`~radis.spectrum.spectrum.Spectrum` object

        Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something
        among ``['radiance', 'radiance_noslit', 'absorbance', etc...]``

        Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method
        to plot it

        See [1]_ to get an overview of all Spectrum methods

        Notes
        -----

        Calculate line strenghts correcting the CDSD reference one. Then call
        the main routine that sums over all lines


        References
        ----------

        .. [1] RADIS doc: `Spectrum how to? <http://radis.readthedocs.io/en/latest/#the-spectrum-class>`__


        See Also
        --------

        :meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`

        '''

        try:

            # %% Preprocessing
            # --------------------------------------------------------------------

            # Check inputs
            if not self.input.self_absorption:
                raise ValueError('Use non_eq_spectrum(Tgas, Tgas) to calculate spectra ' +
                                 'without self_absorption')

            # update defaults
            if path_length is not None:
                self.input.path_length = path_length
            if mole_fraction is not None:
                self.input.mole_fraction = mole_fraction
            if pressure is not None:
                self.input.pressure_mbar = pressure*1e3
            if not is_float(Tgas):
                raise ValueError(
                    'Tgas should be float. Use ParallelFactory for multiple cases')
            self.input.rot_distribution = 'boltzmann'  # equilibrium
            self.input.vib_distribution = 'boltzmann'  # equilibrium

            # Get temperatures
            self.input.Tgas = Tgas
            self.input.Tvib = Tgas  # just for info
            self.input.Trot = Tgas  # just for info

            # Init variables
            pressure_mbar = self.input.pressure_mbar
            mole_fraction = self.input.mole_fraction
            path_length = self.input.path_length
            verbose = self.verbose

            # Check variables
            self._check_inputs(mole_fraction)

            # Retrieve Spectrum from database if it exists
            if self.autoretrievedatabase:
                s = self._retrieve_from_database()
                if s is not None:
                    return s    # exit function

            # %% Start
            # --------------------------------------------------------------------

            t0 = time()
            if verbose:
                self.print_conditions('Calculating Equilibrium Spectrum')

            # Check database, reset populations, create line dataframe to be scaled
            # --------------------------------------------------------------------
            self._check_line_databank()
            self._reinitialize()   # creates scaled dataframe df1 from df0

            # --------------------------------------------------------------------

            # First calculate the linestrength at given temperature
            self._calc_linestrength_eq(Tgas)
            self._cutoff_linestrength()

            # ----------------------------------------------------------------------

            # Calculate line shift
            self._calc_lineshift()

            # ----------------------------------------------------------------------
            # Line broadening

            # ... calculate broadening  FWHM
            self._calc_broadening_FWHM()

            # ... find weak lines and calculate semi-continuum (optional)
            I_continuum = self._calculate_pseudo_continuum()

            # ... apply lineshape and get absorption coefficient
            # ... (this is the performance bottleneck)
            wavenumber, abscoeff_v = self._calc_broadening()
            #    :         :
            #   cm-1    1/(#.cm-2)

            # ... add semi-continuum (optional)
            abscoeff_v = self._add_pseudo_continuum(abscoeff_v, I_continuum)

            # Calculate output quantities
            # ----------------------------------------------------------------------

            # incorporate density of molecules (see equation (A.16) )
            density = mole_fraction*((pressure_mbar*100)/(k_b*Tgas))*1e-6
            #  :
            # (#/cm3)

            abscoeff = abscoeff_v * density      # cm-1

            # ... # TODO: if the code is extended to multi-species, then density has to be added
            # ... before lineshape broadening (as it would not be constant for all species)

            # get absorbance (technically it's the optical depth `tau`,
            #                absorbance `A` being `A = tau/ln(10)` )
            absorbance = abscoeff*path_length

            # Generate output quantities
            transmittance_noslit = exp(-absorbance)
            emissivity_noslit = 1 - transmittance_noslit
            radiance_noslit = calc_radiance(wavenumber, emissivity_noslit, Tgas,
                                            unit=self.units['radiance_noslit'])

            # %% Export
            # --------------------------------------------------------------------

            t = round(time() - t0, 2)
            if verbose:
                print('... process done in {0:.1f}s'.format(t))

            # Get conditions
            conditions = self.get_conditions()
            conditions.update({'calculation_time': t,
                               'lines_calculated': self._Nlines_calculated,
                               'lines_cutoff': self._Nlines_cutoff,
                               'lines_in_continuum': self._Nlines_in_continuum,
                               'thermal_equilibrium':True,
                               'neq_version':get_version(add_git_number=False)})

            # Get populations of levels as calculated in RovibrationalPartitionFunctions
            # ... Populations cannot be calculated at equilibrium (needs energies).
            # ... Use SpectrumFactory.non_eq_spectrum
            populations = None

            # Get lines (intensities + populations)
            lines = self.get_lines()

            # Spectral quantities
            quantities = {
                'abscoeff': (wavenumber, abscoeff),
                'absorbance': (wavenumber, absorbance),
                'emissivity_noslit': (wavenumber, emissivity_noslit),
                'transmittance_noslit': (wavenumber, transmittance_noslit),
                # (mW/cm2/sr/nm)
                'radiance_noslit': (wavenumber, radiance_noslit),
            }
            if I_continuum is not None and self._export_continuum:
                quantities.update(
                    {'abscoeff_continuum': (wavenumber, I_continuum*density)})

            # Store results in Spectrum class
            s = Spectrum(
                quantities=quantities,
                conditions=conditions,
                populations=populations,
                lines=lines,
                units=self.units,
                cond_units=self.cond_units,
                waveunit=self.params.waveunit,  # cm-1
                # dont check input (much faster, and Spectrum
                warnings=False,
                # is freshly baken so probably in a good format
                name=name,
            )

            # update database if asked so
            if self.autoupdatedatabase:
                self.SpecDatabase.add(s, if_exists_then='increment') 
                # Tvib=Trot=Tgas... but this way names in a database
                # generated with eq_spectrum are consistent with names
                # in one generated with non_eq_spectrum

            return s

        except:
            # An error occured: clean before crashing
            self._clean_temp_file()
            raise

    def non_eq_spectrum(self, Tvib, Trot, Ttrans=None, mole_fraction=None, 
                        path_length=None, pressure=None,
                        vib_distribution='boltzmann', rot_distribution='boltzmann',
                        overpopulation=None, name=None):
        ''' Calculate emission spectrum in non-equilibrium case. Calculates
        absorption with broadened linestrength and emission with broadened
        Einstein coefficient.

        Parameters
        ----------

        Tvib: float
            vibrational temperature [K]
            can be a tuple of float for the special case of more-than-diatomic
            molecules (e.g: CO2)

        Trot: float
            rotational temperature [K]

        Ttrans: float
            translational temperature [K]. If None, translational temperature is
            taken as rotational temperature (valid at 1 atm for times above ~ 2ns
            which is the RT characteristic time)

        mole_fraction: float
            database species mole fraction. If None, Factory mole fraction is used.

        path_length: float
            slab size (cm). If None, Factory mole fraction is used.

        pressure: float
            pressure (bar). If None, the default Factory pressure is used.

        vib_distribution: ``'boltzmann'``, ``'treanor'``
            vibrational distribution

        rot_distribution: ``'boltzmann'``
            rotational distribution

        overpopulation: dict, or ``None``
            add overpopulation factors for given levels:

            >>> {level:overpopulation_factor}

        name: str
            case name (useful in batch)

        Returns
        -------

        Returns a :class:`~radis.spectrum.spectrum.Spectrum` object

        Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something
        among ``['radiance', 'radiance_noslit', 'absorbance', etc...]``

        Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method
        to plot it

        See [1]_ to get an overview of all Spectrum methods

        References
        ----------

        .. [1] RADIS doc: `Spectrum how to? <http://radis.readthedocs.io/en/latest/#the-spectrum-class>`__

        See Also
        --------

        :meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`
        :meth:`~radis.lbl.factory.SpectrumFactory.optically_thin_power`

        '''

        try:

            # %% Preprocessing
            # --------------------------------------------------------------------

            # check inputs, update defaults
            if path_length is not None:
                self.input.path_length = path_length
            if mole_fraction is not None:
                self.input.mole_fraction = mole_fraction
            if pressure is not None:
                self.input.pressure_mbar = pressure*1e3
            if not (is_float(Tvib) or isinstance(Tvib, tuple)):
                raise TypeError('Tvib should be float, or tuple (got {0})'.format(type(Tvib)) +
                                'For parallel processing use ParallelFactory with a ' +
                                'list of float or a list of tuple')
            singleTvibmode = is_float(Tvib)
            if not is_float(Trot):
                raise ValueError(
                    'Trot should be float. Use ParallelFactory for multiple cases')
            if overpopulation is None:
                overpopulation = {}
            assert vib_distribution in ['boltzmann', 'treanor']
            assert rot_distribution in ['boltzmann']
            self.input.overpopulation = overpopulation
            self.input.rot_distribution = rot_distribution
            self.input.vib_distribution = vib_distribution

            # Get translational temperature
            Tgas = Ttrans
            if Tgas is None:
                Tgas = Trot     # assuming Ttrans = Trot
            self.input.Tgas = Tgas
            self.input.Tvib = Tvib
            self.input.Trot = Trot

            # Init variables
            path_length = self.input.path_length
            mole_fraction = self.input.mole_fraction
            pressure_mbar = self.input.pressure_mbar
            verbose = self.verbose

            # Check variables
            self._check_inputs(mole_fraction)

            # Retrieve Spectrum from database if it exists
            if self.autoretrievedatabase:
                s = self._retrieve_from_database()
                if s is not None:
                    return s  # exit function

            # %% Start
            # --------------------------------------------------------------------

            t0 = time()
            if verbose:
                self.print_conditions('Calculating Non-Equilibrium Spectrum')

            # Check line database and parameters, reset populations and scaled line dataframe
            # ----------
            self._check_line_databank()
            self._check_noneq_parameters(vib_distribution, singleTvibmode)
            self._reinitialize()   # creates scaled dataframe df1 from df0

            # ----------------------------------------------------------------------
            # Calculate Populations, Linestrength and Emission Integral
            if singleTvibmode:
                self._calc_populations_noneq(Tvib, Trot,
                                             vib_distribution=vib_distribution,
                                             rot_distribution=rot_distribution,
                                             overpopulation=overpopulation)
            else:
                self._calc_populations_noneq_multiTvib(Tvib, Trot,
                                                       vib_distribution=vib_distribution,
                                                       rot_distribution=rot_distribution,
                                                       overpopulation=overpopulation)
            
#            assert not (self.df1.Qref.is_monotonic and self.df1.wav.is_monotonic)
            
            self._calc_linestrength_noneq()
            self._calc_emission_integral()

#            assert not (self.df1.Qref.is_monotonic and self.df1.wav.is_monotonic)
            
            # ----------------------------------------------------------------------
            # Cutoff linestrength
            self._cutoff_linestrength()

            # ----------------------------------------------------------------------

            # Calculate lineshift
            self._calc_lineshift()

            # ----------------------------------------------------------------------

            # Line broadening

            # ... calculate broadening  FWHM
            self._calc_broadening_FWHM()

            # ... find weak lines and calculate semi-continuum (optional)
            k_continuum, j_continuum = self._calculate_pseudo_continuum(noneq=True)

            # ... apply lineshape and get absorption coefficient
            # ... (this is the performance bottleneck)
            wavenumber, abscoeff_v, emisscoeff_v = self._calc_broadening_noneq()
            #    :         :            :
            #   cm-1    1/(#.cm-2)   mW/sr/cm_1

            # ... add semi-continuum (optional)
            abscoeff_v = self._add_pseudo_continuum(abscoeff_v, k_continuum)
            emisscoeff_v = self._add_pseudo_continuum(emisscoeff_v, j_continuum)

            # Calculate output quantities
            # ----------------------------------------------------------------------

            # incorporate density of molecules (see Rothman 1996 equation (A.16) )
            density = mole_fraction*((pressure_mbar*100)/(k_b*Tgas))*1e-6
            #  :
            # (#/cm3)

            abscoeff = abscoeff_v * density                     # cm-1
            emisscoeff = emisscoeff_v * density                 # mW/sr/cm3/cm_1

            # ... # TODO: if the code is extended to multi-species, then density has to be added
            # ... before lineshape broadening (as it would not be constant for all species)

            # get absorbance (technically it's the optical depth `tau`,
            #                absorbance `A` being `A = tau/ln(10)` )

            # Generate output quantities
            absorbance = abscoeff*path_length           # (adim)
            transmittance_noslit = exp(-absorbance)

            if self.input.self_absorption:
                # Analytical output of computing RTE over a single slab of constant
                # emissivity and absorption coefficient
                b = (abscoeff == 0)  # optically thin mask
                radiance_noslit = np.zeros_like(emisscoeff)
                radiance_noslit[~b] = emisscoeff[~b] / \
                    abscoeff[~b]*(1-transmittance_noslit[~b])
                radiance_noslit[b] = emisscoeff[b]*path_length
            else:
                # Note that for k -> 0,
                radiance_noslit = emisscoeff*path_length    # (mW/sr/cm2/cm_1)

            # Convert `radiance_noslit` from (mW/sr/cm2/cm_1) to (mW/sr/cm2/nm)
            radiance_noslit = convert_rad2nm(radiance_noslit, wavenumber,
                                             'mW/sr/cm2/cm_1', 'mW/sr/cm2/nm')
            # Convert 'emisscoeff' from (mW/sr/cm3/cm_1) to (mW/sr/cm3/nm)
            emisscoeff = convert_emi2nm(emisscoeff, wavenumber,
                                        'mW/sr/cm3/cm_1', 'mW/sr/cm3/nm')

            # Note: emissivity not defined under non equilibrium

            # %% Export
            # ----------------------------------------------------------------------

            t = round(time() - t0, 2)
            if verbose:
                print('... process done in {0:.1f}s'.format(t))

            # Get conditions
            conditions = self.get_conditions()
            conditions.update({'calculation_time': t,
                               'lines_calculated': self._Nlines_calculated,
                               'lines_cutoff': self._Nlines_cutoff,
                               'lines_in_continuum': self._Nlines_in_continuum,
                               'thermal_equilibrium': False, # dont even try to guess if it's at equilibrium
                               'neq_version':get_version(add_git_number=False)})

            # Get populations of levels as calculated in RovibrationalPartitionFunctions
            populations = self.get_populations(self.misc.export_populations)

            # Get lines (intensities + populations)
            lines = self.get_lines()

            # Spectral quantities
            quantities = {
                'abscoeff': (wavenumber, abscoeff),
                'absorbance': (wavenumber, absorbance),
                # (mW/cm3/sr/nm)
                'emisscoeff': (wavenumber, emisscoeff),
                'transmittance_noslit': (wavenumber, transmittance_noslit),
                # (mW/cm2/sr/nm)
                'radiance_noslit': (wavenumber, radiance_noslit),
            }
            if k_continuum is not None and self._export_continuum:
                quantities.update(
                    {'abscoeff_continuum': (wavenumber, k_continuum*density),
                     'emisscoeff_continuum': (wavenumber, j_continuum*density)})

            # Store results in Spectrum class
            s = Spectrum(
                quantities=quantities,
                conditions=conditions,
                populations=populations,
                lines=lines,
                units=self.units,
                cond_units=self.cond_units,
                waveunit=self.params.waveunit,  # cm-1
                # dont check input (much faster, and Spectrum
                warnings=False,
                # is freshly baken so probably in a good format
                name=name,
            )

            # update database if asked so
            if self.autoupdatedatabase:
                self.SpecDatabase.add(
                    s, add_info=['Tvib', 'Trot'], add_date='%Y%m%d', if_exists_then='increment')

            return s

        except:
            # An error occured: clean before crashing
            self._clean_temp_file()
            raise

    def optically_thin_power(self, Tgas=None, Tvib=None, Trot=None, Ttrans=None,
                             mole_fraction=None, path_length=None, unit='mW/cm2/sr'):
        ''' Calculate total power emitted in equilibrium or non-equilibrium case
        in the optically thin approximation: it sums all emission integral over
        the total spectral range.

        .. warning::
        
            this is a fast implementation that doesnt take into account
            the contribution of lines outside the given spectral range. It is valid for spectral ranges
            surrounded by no lines, and spectral ranges much broaded than the typical
            line broadening (~ 1-10 cm-1 in the infrared)

        If what you're looking for is an accurate simulation on a narrow spectral range
        you better calculate the spectrum (that does take all of that into account)
        and integrate it with :py:meth:`~radis.spectrum.spectrum.Spectrum.get_power`

        Parameters
        ----------

        Tgas: float
            equilibrium temperature [K]
            If doing a non equilibrium case it should be None. Use Ttrans for
            translational temperature

        Tvib: float
            vibrational temperature [K]

        Trot: float
            rotational temperature [K]

        Ttrans: float
            translational temperature [K]. If None, translational temperature is
            taken as rotational temperature (valid at 1 atm for times above ~ 2ns
            which is the RT characteristic time)

        mole_fraction: float
            database species mole fraction. If None, Factory mole fraction is used.

        path_length: float
            slab size (cm). If None, Factory mole fraction is used.

        unit: str
            output unit. Default ``'mW/cm2/sr'``

        Returns
        -------

        Returns total power density in mW/cm2/sr (unless different unit is chosen),
        see ``unit=``.


        See Also
        --------

        :py:meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`, 
        :py:meth:`~radis.spectrum.spectrum.Spectrum.get_power`, 
        :py:meth:`~radis.spectrum.spectrum.Spectrum.get_integral`
        
        '''

        try:

            # Check inputs

            # ... temperatures

            if Tgas is None and Trot is None:
                raise ValueError('Choose either Tgas (equilibrium) or Tvib / Trot ' +
                                 '. Ttrans (non equilibrium)')

            non_eq_mode = (Tgas is None)

            if Tvib is None and Trot is not None or Tvib is not None and Trot is None:
                raise ValueError('Choose both Tvib and Trot')

            if non_eq_mode and Ttrans is None:
                Ttrans = Trot

            # update defaults
            if path_length is not None:
                self.input.path_length = path_length
            if mole_fraction is not None:
                self.input.mole_fraction = mole_fraction

            # Get translational temperature
            if Tgas is None:
                Tgas = Ttrans
            self.input.Tgas = Tgas
            self.input.Tvib = Tvib
            self.input.Trot = Trot

            # Init variables
            path_length = self.input.path_length
            mole_fraction = self.input.mole_fraction
            pressure_mbar = self.input.pressure_mbar
            verbose = self.verbose

            # Make sure database is loaded
            if self.df0 is None:
                if not self.save_memory:
                    raise AttributeError(
                        'Load databank first (.load_databank())')
                else:
                    self._reload_databank()

            if non_eq_mode:
                # Make sure database has pre-computed non equilibrium quantities
                # (Evib, Erot, etc.)
                try:
                    self.df0['Evib']
                except KeyError:
                    self._calc_noneq_parameters()

                try:
                    self.df0['Aul']
                except KeyError:
                    self._calc_weighted_trans_moment()
                    self._calc_einstein_coefficients()

            # %% Start
            # ----------------------------------------------------------------------

            # Print conditions
            if verbose:
                self.print_conditions(
                    'Calculating Radiative Power (optically thin approximation)')

            # Calculate power
            # ---------------------------------------------------

            self._reinitialize()   # creates scaled dataframe df1 from df0

            # ----------------------------------------------------------------------
            # Calculate Populations and Emission Integral
            # (Note: Emission Integral is non canonical quantity, equivalent to
            #  Linestrength for absorption)
            if non_eq_mode:
                self._calc_populations_noneq(Tvib, Trot)
            else:
                self._calc_populations_eq(Tgas)
                self.df1['Aul'] = self.df1.A  # update einstein coefficients
            self._calc_emission_integral()

    #        # ----------------------------------------------------------------------
    #        # Cutoff linestrength  (note that cuting linestrength doesnt make this
    #        # process faster here, but we still give this option to be consistent
    #        # with spectra)
    #        self._cutoff_linestrength()

            # ----------------------------------------------------------------------

            # Sum over all emission integrals (in the valid range)
            b = (self.df1.wav > self.input.wavenum_min) & (
                self.df1.wav < self.input.wavenum_max)
            P = self.df1['Ei'][b].sum()          # Ei  >> (mW/sr)

            # incorporate density of molecules (see equation (A.16) )
            density = mole_fraction*((pressure_mbar*100)/(k_b*Tgas))*1e-6
            Pv = P * density                     # (mW/sr/cm3)

            # Optically thin case (no self absorption):
            Ptot = Pv*path_length    # (mW/sr/cm2)

            return conv2(Ptot, 'mW/cm2/sr', unit)

        except:
            # An error occured: clean before crashing
            self._clean_temp_file()
            raise

# %% ======================================================================
# EXTRA FUNCTIONS
# ---------------------
#
# _generate_wavenumber_range
# _generate_broadening_range
#
# XXX =====================================================================


def _generate_wavenumber_range(wavenum_min, wavenum_max, wstep, broadening_max_width):
    ''' define waverange vectors, with ``wavenumber`` the ouput spectral range
    and ``wavenumber_calc`` the spectral range used for calculation, that includes
    neighbour lines within ``broadening_max_width`` distance

    Parameters
    ----------

    wavenum_min, wavenum_max: float
        wavenumber range limits (cm-1)

    wstep: float
        wavenumber step (cm-1)

    broadening_max_width: float
        wavenumber full width of broadening calculation: used to define which
        neighbour lines shall be included in the calculation

    Returns
    -------

    wavenumber: numpy array
        an evenly spaced array between ``wavenum_min`` and ``wavenum_max`` with
        a spacing of ``wstep``

    wavenumber_calc: numpy array
        an evenly spaced array between ``wavenum_min-broadening_max_width/2`` and
        ``wavenum_max+broadening_max_width/2`` with a spacing of ``wstep``

    '''

    assert wavenum_min < wavenum_max

    # Output range
    # generate the vector of wavenumbers (shape M)
    wavenumber = arange(wavenum_min, wavenum_max+wstep, wstep)

    # Calculation range
    wavenum_min_calc = wavenumber[0] - broadening_max_width/2     # cm-1
    wavenum_max_calc = wavenumber[-1] + broadening_max_width/2     # cm-1
    w_out_of_range_left = arange(
        wavenumber[0]-wstep, wavenum_min_calc-wstep, -wstep)[::-1]
    w_out_of_range_right = arange(
        wavenumber[-1]+wstep, wavenum_max_calc+wstep, wstep)
#    w_out_of_range_right = -w_out_of_range_left + wavenumber[-1] + wstep - w_out_of_range_left[1]
    wavenumber_calc = np.hstack(
        (w_out_of_range_left, wavenumber, w_out_of_range_right))
    
    # deal with rounding errors: 1 side may have 1 more point
    if len(w_out_of_range_left) > len(w_out_of_range_right):
        w_out_of_range_left = w_out_of_range_left[1:]
    elif len(w_out_of_range_left) < len(w_out_of_range_right):
        w_out_of_range_right = w_out_of_range_right[:-1]

    assert len(w_out_of_range_left) == len(w_out_of_range_right)

    return wavenumber, wavenumber_calc


def _generate_broadening_range(wstep, broadening_max_width):
    ''' Generate array on which to compute line broadening

    Parameters
    ----------

    wstep: float
        wavenumber step (cm-1)

    broadening_max_width: float
        wavenumber full width of broadening calculation: used to define which
        neighbour lines shall be included in the calculation

    Returns
    -------

    wbroad_centered: numpy array
        an evenly spaced array, of odd-parity length, centered on 0, and of width
        ``broadening_max_width``

    '''

    # create a broadening array, on which lineshape will be calculated.
    # Odd number is important
    wbroad_centered = np.hstack((-arange(wstep, 0.5*broadening_max_width+wstep, wstep)[::-1],
                                 [0],
                                 arange(wstep, 0.5*broadening_max_width+wstep, wstep)))

    assert len(wbroad_centered) % 2 == 1

    return wbroad_centered

# %% Test

# --------------------------
if __name__ == '__main__':
    from radis.test.lbl.test_factory import _run_testcases
    print('Testing factory:', _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Summary
-------

Functions to compare and plot comparison of :class:`~radis.spectrum.spectrum.Spectrum` 
objects


Routine Listings
----------------

- :func:`~radis.spectrum.compare.get_diff`
- :func:`~radis.spectrum.compare.get_distance`
- :func:`~radis.spectrum.compare.get_ratio`
- :func:`~radis.spectrum.compare.get_residual`
- :func:`~radis.spectrum.compare.get_residual_integral`
- :func:`~radis.spectrum.compare.plot_diff`


-------------------------------------------------------------------------------



"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.misc.arrays import array_allclose
from radis.misc.curve import curve_substract, curve_distance, curve_divide
from radis.spectrum.spectrum import Spectrum, make_up, cast_waveunit, is_spectrum
from radis.misc.basics import compare_lists, compare_dict
from six import string_types

import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.widgets import MultiCursor
import numpy as np
from publib import set_style, fix_style
from warnings import warn
from six.moves import range
from six.moves import zip


# %% ======================================================================
# External functions
# ----------------
# XXX =====================================================================


def get_diff(s1, s2, var, wunit='default', Iunit='default', medium='default',
             resample=True, diff_window=0):
    # type: (Spectrum, Spectrum, str, str, str, str, bool, int) -> np.array, np.array
    ''' Get the difference between 2 spectra
    Basically returns w1, I1 - I2 where (w1, I1) and (w2, I2) are the values of
    s1 and s2 for variable var. (w2, I2) is linearly interpolated if needed.

        .. math::

            dI = I_1 - I_2


    Parameters    
    ----------

    s1, s2: Spectrum objects
        2 spectra to compare.

    var: str
        spectral quantity (ex: 'radiance', 'transmittance'...)

    wunit: 'nm', 'cm-1'
        waveunit to compare in

    Iunit: str
        if ``'default'`` use s1 unit for variable var

    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)

    Other Parameters
    ----------------
    
    resample: bool
        if not ``True``, wavelength must be equals. Else, resample ``s2`` on
        ``s1`` if needed.
    
    diff_window: int
        If non 0, calculates diff by offsetting s1 by ``diff_window`` number of
        units on either side, and returns the minimum. Kinda compensates for experimental
        errors on the w axis. Default 0. (look up code to understand...)

    Returns    
    -------

    w1, Idiff: array
        difference interpolated on the second range(order?)
    
    Notes
    -----
    
    Uses :func:`~radis.misc.curve.curve_substract` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''
    
    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit,
                                   medium=medium, assert_same_wavelength=not resample)

    # basically w1, I1 - I2 (on same range)
    w1, Idiff = curve_substract(w1, I1, w2, I2)
    if diff_window:
        # allow fluctuation from diff_window units. Kinda compensates 
        # for experimental errors on x-axis
        diff_list = []
        I2_interp = I1 - Idiff
        for i in range(-diff_window, diff_window+1):
            diff_list.append(I2_interp-np.roll(I1, i))
        # get minimum in abs value
        diff_list = np.array(diff_list)
        b = np.abs(diff_list).argmin(axis=0)
        Idiff = np.choose(b, diff_list)   # keeps sign
    return w1, Idiff
        
        
        


def get_ratio(s1, s2, var, 
              wunit='default', Iunit='default', medium='default', resample=True):
    # type: (Spectrum, Spectrum, str, str, str, str, bool) -> np.array, np.array
    ''' Get the ratio between two spectra
    Basically returns w1, I1 / I2 where (w1, I1) and (w2, I2) are the values of
    s1 and s2 for variable var. (w2, I2) is linearly interpolated if needed.

        .. math::

            R = I_1 / I_2


    Parameters    
    ----------

    s1, s2: Spectrum objects

    var: str
        spectral quantity 

    wunit: 'nm', 'cm-1'
        waveunit to compare in

    Iunit: str
        if ``'default'`` use s1 unit for variable var

    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)

    Notes
    -----
    
    Uses :func:`~radis.misc.curve.curve_divide` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_distance`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with` 
    

    '''
    
    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit,
                                   medium=medium, assert_same_wavelength=not resample)

    # basically w1, I1 - I2 (on same range)
    return curve_divide(w1, I1, w2, I2)


def get_distance(s1, s2, var, 
                 wunit='default', Iunit='default', medium='default', resample=True):
    # type: (Spectrum, Spectrum, str, str, str, str, bool) -> np.array, np.array
    ''' Get a regularized Euclidian distance between two spectra ``s1`` and ``s2`` 

    This regularized Euclidian distance minimizes the effect of a small shift in 
    wavelength between the two spectra

    .. math::

        D(w_1)[i] = \sqrt{ \sum_j (\hat{I_1}[i]  - \hat{I_2}[j] )^2 + (\hat{w_1}[i] - \hat{w_2}[j])^2}
        
    Where values are normalized as:
        
    .. math::

        \hat{A} = \\frac{A}{max(A) - min(A)}

    If waveranges dont match, ``s2`` is interpolated over ``s1``.

    .. warning :: 
        
        This is a distance on both the waverange and the intensity axis. 
        It may be used to compensate for a small offset in your experimental 
        spectrum (due to wavelength calibration, for instance) but can lead
        to wrong fits easily. Plus, it is very cost-intensive! Better use 
        :func:`~radis.spectrum.compare.get_residual` for an automatized procedure. 


    Parameters    
    ----------

    s1, s2: Spectrum objects

    var: str
        spectral quantity 

    wunit: 'nm', 'cm-1'
        waveunit to compare in

    Iunit: str
        if ``'default'`` use s1 unit for variable var

    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)
    
    Notes
    -----
    
    Uses :func:`~radis.misc.curve.curve_distance` internally

    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with` 

    '''
    # TODO: normalize with Imax, wmax 

    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit,
                                   medium=medium, assert_same_wavelength=not resample)

    # euclidian distance from w1, I1
    return curve_distance(w1, I1, w2, I2, discard_out_of_bounds=True)


def get_residual(s1, s2, var, 
                 norm='L2', ignore_nan=False, diff_window=0,
                 normalize=False, normalize_how='max'):
    # type: (Spectrum, Spectrum, str, bool, int) -> np.array, np.array
    ''' Returns L2 norm of ``s1`` and ``s2``

    For ``I1``, ``I2``, the values of variable ``var`` in ``s1`` and ``s2``, 
    respectively, residual is calculated as:
        
    For ``L2`` norm:

        .. math::
    
            res = \\frac{\\sqrt{\\sum_i {(s_1[i]-s_2[i])^2}}}{N}.

    For ``L1`` norm:
            
        .. math::
    
            res = \\frac{\\sqrt{\\sum_i {|s_1[i]-s_2[i]|}}}{N}.


    Parameters    
    ----------

    s1, s2: :class:`~radis.spectrum.spectrum.Spectrum` objects
        if not on the same range, ``s2`` is resampled on ``s1``.

    var: str
        spectral quantity

    norm: 'L2', 'L1'
        which norm to use 

    Other Parameters
    ----------------

    ignore_nan: boolean
        if ``True``, ignore nan in the difference between s1 and s2 (ex: out of bound)
        when calculating residual. Default ``False``. Note: ``get_residual`` will still 
        fail if there are nan in initial Spectrum. 

    normalize: bool, or tuple
        if ``True``, normalize the two spectra before calculating the residual. 
        If a tuple (ex: ``(4168, 4180)``), normalize on this range only. The unit
        is that of the first Spectrum. Ex::
            
            s_exp   # in 'nm'
            s_calc  # in 'cm-1'
            get_residual(s_exp, s_calc, normalize=(4178, 4180))  # interpreted as 'nm'

    normalize_how: 'max', 'area'
        how to normalize. 'max' is the default but may not be suited for very
        noisy experimental spectra. 

    Notes
    -----

    0 values for I1 yield nans except if I2 = I1 = 0

    when s1 and s2 dont have the size wavespace range, they are automatically
    resampled through get_diff on 's1' range 

    Implementation of ``L2`` norm::

        np.sqrt((dI**2).sum())/len(dI)
    
    Implementation of ``L1`` norm::
        
        np.abs(dI).sum()/len(dI)

    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.plot_diff`, 
    :func:`~radis.spectrum.compare.get_residual_integral`, 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''

    if normalize:
        from radis.spectrum.operations import multiply
        if isinstance(normalize, tuple):
            wmin, wmax = normalize
            w1, I1 = s1.get(var, copy=False)  # (faster not to copy)
            b = (w1 > wmin) & (w1 < wmax)
            if normalize_how == 'max':
                norm1 = I1[b].max()
            elif normalize_how == 'area':
                norm1 = np.abs(np.trapz(I1[b], w1[b]))
            # now normalize s2. Ensure we use the same unit system!
            w2, I2 = s2.get(var, Iunit=s1.units[var], wunit=s1.get_waveunit())
            b = (w2 > wmin) & (w2 < wmax)
            if normalize_how == 'max':
                norm2 = I2[b].max()
            elif normalize_how == 'area':
                norm2 = np.abs(np.trapz(I2[b], w2[b]))
            s1 = multiply(s1, 1/norm1, var=var)
            s2 = multiply(s2, 1/norm2, var=var)
        else:
            if normalize_how == 'max':
                norm1 = s1.get(var, copy=False)[1].max()
                norm2 = s2.get(var)[1].max()
            elif normalize_how == 'area':
                norm1 = s1.get_integral(var)
                norm2 = s2.get_integral(var, wunit=s1.get_waveunit(), Iunit=s1.units[var])
            # Ensure we use the same unit system!
            s1 = multiply(s1, 1/norm1, var=var)
            s2 = multiply(s2, 1/norm2, var=var)

    # mask for 0
    wdiff, dI = get_diff(s1, s2, var, resample=True, diff_window=diff_window)

    if ignore_nan:
        b = np.isnan(dI)
        wdiff, dI = wdiff[~b], dI[~b]
    warningText = 'NaN output in residual. You should use "ignore_nan=True". Read the help.'
    if norm == 'L2':
        output = np.sqrt((dI**2).sum())/len(dI)
        if np.isnan(output):
            warn(warningText, UserWarning)
        return output
    elif norm == 'L1':
        output = (np.abs(dI)).sum()/len(dI)
        if np.isnan(output):
            warn.warning(warningText, UserWarning)
        return output
    else:
        raise ValueError('unexpected value for norm')


def get_residual_integral(s1, s2, var, 
                          ignore_nan=False):
    # type: (Spectrum, Spectrum, str, bool) -> float
    ''' Returns integral of the difference between two spectra s1 and s2, 
    relatively to the integral of spectrum s1 
    
    Compared to :func:`~radis.spectrum.compare.get_residual`, this tends to 
    cancel the effect of the gaussian noise of an experimental spectrum. 
    
    ::
    
        res = trapz(I2-I1, w1) / trapz(I1, w1)

    Note: when the considered variable is ``transmittance`` or ``transmittance_noslit``, 
    the *upper* integral is used (up to 1) to normalize the integral difference
    
    ::
    
        res = trapz(I2-I1, w1) / trapz(1-I1, w1)

    Parameters
    ----------

    s1, s2: Spectrum objects

    var: str
        spectral quantity

    Other Parameters
    ----------------

    ignore_nan: boolean
        if ``True``, ignore nan in the difference between s1 and s2 (ex: out of bound)
        when calculating residual. Default ``False``. Note: ``get_residual_integral``
        will still fail if there are nan in initial Spectrum. 

    Notes
    -----

    For I1, I2, the values of 'var' in s1 and s2, respectively, residual
    is calculated as::

        res = trapz(I2-I1, w1) / trapz(I1, w1)

    0 values for I1 yield nans except if I2 = I1 = 0

    when s1 and s2 dont have the size wavespace range, they are automatically
    resampled through get_diff on 's1' range 


    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.get_residual`, 
    :func:`~radis.spectrum.compare.plot_diff`,
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''

    w1, I1, _, _ = _get_defaults(s1, s2, var)
    # mask for 0
    wdiff, dI = get_diff(s1, s2, var, resample=True)

    if ignore_nan:
        b = np.isnan(dI)
        wdiff, dI = wdiff[~b], dI[~b]
        b = np.isnan(I1)
        w1, I1 = w1[~b], I1[~b]
        
    if var in ['transmittance', 'transmittance_noslit']:
        norm = 1 - np.trapz(I1, w1)
    else:
        norm = np.trapz(I1, w1)

    return np.abs(np.trapz(dI, wdiff)) / norm

#    b1 = (I_avg == 0)
#    b2 = (dI == 0)

#    with catch_warnings():
#        filterwarnings('ignore', 'invalid value encountered in true_divide')
#        res = dI/I_avg

#    res[b1*b2] = 0             # if both I_avg and dI = 0, solve nan as 0

#    return np.sqrt((res**2).sum())/len(res)
#    return res.mean()


def _get_defaults(s1, s2, var, 
                  wunit='default', Iunit='default', medium='default',
                  assert_same_wavelength=False):
    # type: (Spectrum, Spectrum, str, str, str, bool) -> (np.array, np.array, np.array, np.array)
    ''' Returns w1, I1, w2, I2  in the same waveunit, unit and medium.
    
    See get_distance, get_diff for more information '''

    # Check inputs, get defaults
    # ----
    if Iunit == 'default':
        try:
            Iunit = s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError('Iunit not defined in spectrum for variable {0}'.format(var))
    # Format units
    if wunit == 'default':
        wunit = s1.get_waveunit()
    wunit = cast_waveunit(wunit)
    if medium == 'default':
        medium = s1.conditions.get('medium', None)

    # Get data
    # ----
    w1, I1 = s1.get(var, wunit=wunit, Iunit=Iunit, medium=medium)
    w2, I2 = s2.get(var, wunit=wunit, Iunit=Iunit, medium=medium)
    
    if assert_same_wavelength:
        if not array_allclose(w1, w2):
            raise AssertionError(
                'Wavespace are not the same: use Spectrum.resample()')

    return w1, I1, w2, I2


def plot_diff(s1, s2, var=None, 
              wunit='default', Iunit='default', medium='default',
              resample=True, method='diff', diff_window=0, show_points=False,
              label1=None, label2=None, figsize=None, title=None, nfig=None,
              normalize=False, verbose=True, save=False, show=True,
              show_residual=False, lw_multiplier=1, diff_scale_multiplier=1,
              discard_centile=0):
    ''' Plot two spectra, and the difference between them. ``method=`` allows
    you to plot the absolute difference, ratio, or both. 

    If waveranges dont match, ``s2`` is interpolated over ``s1``. 


    Parameters    
    ----------

    s1, s2: Spectrum objects

    var: str, or None
        spectral quantity to plot (ex: ``'abscoeff'``). If None, 
        plot the first one in the Spectrum from ``'radiance'``, 
        ``'radiance_noslit'``, ``'transmittance'``, etc.

    wunit: ``'default'``, ``'nm'``, ``'cm-1'``
        if ``'default'``, use first spectrum wunit

    Iunit: str
        if ``'default'``, use first spectrum unit

    medium: ``'air'``, ``'vacuum'``, ``'default'``
        if ``'default'``, use first spectrum propagating medium

    method: ``'distance'``, ``'diff'``, ``'ratio'``, or list of them.
        If ``'diff'``, plot difference of the two spectra.
        If ``'distance'``, plot Euclidian distance (note that units are meaningless then)
        If ``'ratio'``, plot ratio of two spectra
        Default ``'diff'``.

        .. warning::
            with ``'distance'``, calculation scales as ~N^2 with N the number
            of points in a spectrum (against ~N with ``'diff'``). This can quickly 
            override all memory.
            
        Can also be a list::
            
            method=['diff', 'ratio']

    normalize: bool
        Normalize the spectra to be ploted 

    Other Parameters
    ----------------

    diff_window: int
        If non 0, calculates diff by offsetting s1 by ``diff_window`` number of
        units on either side, and returns the minimum. Kinda compensates for experimental
        errors on the w axis. Default 0. (look up code to understand...)

    show_points: boolean
        if ``True``, make all points appear with 'o'

    label1, label2: str
        curve names

    figsize
        figure size

    nfig: int, str
        figure number of name

    title: str
        title

    verbose: boolean
        if ``True``, plot stuff such as rescale ratio in normalize mode. Default ``True``

    save: str
        Default is ``False``. By default won't save anything, type the path of the 
        destination if you want to save it (format in the name).
        
    show: Bool
        Default is ``True``. Will show the plots : bad if there are more than 20.
    
    show_residual: bool
        if ``True``, calculates and shows on the graph the residual in L2 norm. 
        See :func:`~radis.spectrum.compare.get_residual`. ``diff_window`` is 
        used in the residual calculation too. ``normalize`` has no effect. 
        
    diff_scale_multiplier: float
        dilate the diff plot scale. Default ``1``
    
    discard_centile: int
        if not ``0``, discard the firsts and lasts centile when setting the limits
        of the diff window. Example::
            
            discard_centile=1     #  --> discards the smallest 1% and largest 1%
            discard_centile=10    #  --> discards the smallest 10% and largest 10%
            
        Useful to remove spikes in a ratio, for instance. 
        Note that this does not change the values of the residual. It's just 
        a plot feature.
        Default ``0`` 
        
    Returns
    -------
    
    fig: figure
        fig
    
    [ax0, ax1]: axes
        spectra and difference axis
    
        
    Examples
    --------
    
    Simple use::
        
        from radis import plot_diff
        plot_diff(s10, s50)                # s10, s50 are two spectra

    Advanced use, plotting the total power in the label, and getting the figure 
    and axes handle to edit them afterwards::

        Punit = 'mW/cm2/sr'
        fig, axes = plot_diff(s10, s50, 'radiance_noslit', figsize=(18,6),
              label1='brd 10 cm-1, P={0:.2f} {1}'.format(s10.get_power(unit=Punit),Punit),
              label2='brd 50 cm-1, P={0:.2f} {1}'.format(s50.get_power(unit=Punit),Punit)
              )
        # modify fig, axes..

    See an example output in :ref:`Compare two Spectra <label_spectrum_howto_compare>`

    See Also
    --------

    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.get_residual`, 
    :meth:`~radis.spectrum.spectrum.compare_with` 

    '''
    
    if (not show) and (not save): #I added this line to avoid calculus in the case there is nothing to do (Minou) 
        if verbose: print('plot_diff : Nothing to do')
        return None, None
    
    #Normal behaviour (Minou)
    # Get defaults
    # ---
    if var is None:    # if nothing is defined, try these first:
        params = s1.get_vars()
        if 'radiance' in params:
            var = 'radiance'
        elif 'radiance_noslit' in params:
            var = 'radiance_noslit'
        elif 'transmittance' in params:
            var = 'transmittance'
        elif 'transmittance_noslit' in params:
            var = 'transmittance_noslit'
        else:
            # or plot the first variable we find
            var = list(params)[0]
            if var.replace('_noslit', '') in params:
                var = var.replace('_noslit', '')
    # ... check variable exist
    if var not in s1.get_vars():
        raise ValueError('{0} not defined in Spectrum {1}. Use one of : {2}'.format(
                var, s1.get_name(), s1.get_vars()))
    if var not in s2.get_vars():
        raise ValueError('{0} not defined in Spectrum {1}. Use one of : {2}'.format(
                var, s2.get_name(), s2.get_vars()))
    if Iunit == 'default':
        try:
            Iunit = s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError('Iunit not defined in spectrum for variable {0}. '.format(var)+\
                           "Cant use default unit. Specify unit in s.units['{0}'].".format(var))
    if wunit == 'default':
        wunit = s1.get_waveunit()
    if medium == 'default':
        medium = s1.conditions.get('medium', None)
        
    if isinstance(method, list):
        methods = method
    else:
        methods = [method]

    for method in methods:
        if diff_window != 0 and method != 'diff':
            raise NotImplementedError('diff_window with method {0}'.format(method))
        
    # Get data
    # ----
    if normalize:
        # copy before modifying directly in spectrum
        s1 = s1.copy()
        s2 = s2.copy()
        w1, I1 = s1.get(var, copy=False)
        w2, I2 = s2.get(var, copy=False)
        I1 /= np.max(I1)
        I2 /= np.max(I2)
        if verbose:
            print(('Rescale factor: '+str(np.max(I1)/np.max(I2))))
    
    def get_wdiff_Idiff():
        wdiffs, Idiffs = [], []
        for method in methods:
            if not normalize:
                if method == 'distance':
                    wdiff, Idiff = get_distance(
                        s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
                elif method == 'diff':
                    wdiff, Idiff = get_diff(
                        s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium, 
                        diff_window=diff_window)
                elif method == 'ratio':
                    wdiff, Idiff = get_ratio(
                        s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
                else:
                    raise ValueError('Unknown comparison method: {0}'.format(method))
                wdiffs.append(wdiff)
                Idiffs.append(Idiff)
            else:
                if method == 'distance':
                    raise ValueError('{0} was not implemented yet for normalized spectra'.format(method))
                elif method == 'diff':
                    wdiff, Idiff = curve_substract(w1, I1, w2, I2)
                elif method == 'ratio':
                    wdiff, Idiff = get_ratio(
                        s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
                else:
                    raise ValueError('Unknown comparison method: {0}'.format(method))
                wdiffs.append(wdiff)
                Idiffs.append(Idiff)
        return wdiffs, Idiffs
    
    wdiffs, Idiffs = get_wdiff_Idiff()

    # Plot
    # ----

    # Format units
    wunit = cast_waveunit(wunit)
    if wunit == 'cm-1':
        xlabel = 'Wavenumber (cm-1)'
    elif wunit == 'nm':
        xlabel = 'Wavelength (nm)'

    # Init figure
    set_style('origin')
    fig = plt.figure(num=nfig, figsize=figsize)
    gs = gridspec.GridSpec(1+len(methods), 1, height_ratios=[3]+[1]*len(methods))
    ax0 = plt.subplot(gs[0])
    ax0.ticklabel_format(useOffset=False)
    ax1 = []
    for i in range(len(methods)):
        ax1i = plt.subplot(gs[i+1])
        ax1i.get_shared_x_axes().join(ax0, ax1i)
        ax1i.ticklabel_format(useOffset=False)
        ax1.append(ax1i)

    # Plotting style
    if show_points:
        style = '-o'
    else:
        style = '-'

    # Get labels and names
    if label1 is None:
        label1 = s1.get_name()
    if label2 is None:
        label2 = s2.get_name()
    # Max label length:
    if len(label1)>80:
        label1 = label1[:78]+'...'
    if len(label2)>80:
        label2 = label2[:78]+'...'
        
    # Plot compared spectra
    if normalize:
        # TODO: add option to norm_on
        ax0.plot(w1, I1, ls=style, color='k', lw=3*lw_multiplier, label=label1)
        ax0.plot(w2, I2, ls=style, color='r', lw=1*lw_multiplier, label=label2)
    else:
        ax0.plot(*s1.get(var, wunit, Iunit, medium),
                 ls=style, color='k', lw=3*lw_multiplier, label=label1)
        ax0.plot(*s2.get(var, wunit, Iunit, medium),
                 ls=style, color='r', lw=1*lw_multiplier, label=label2)

    Iunit = make_up(Iunit)  # cosmetic changes

    ax0.tick_params(labelbottom=False)
    if label1 is not None or label2 is not None:
        ax0.legend(loc='best')

    # Start to 0
    if var in ['radiance_noslit', 'radiance', 'abscoeff', 'absorbance']:
        ax0.set_ylim(bottom=0)

    # plot difference (sorted)
    for ax1i, wdiff, Idiff in zip(ax1, wdiffs, Idiffs):
        b = np.argsort(wdiff)
        ax1i.plot(wdiff[b], Idiff[b], style, color='k', lw=1*lw_multiplier)
    
    # Write labels
    ax1[-1].set_xlabel(make_up(xlabel))
    if normalize:
        fig.text(0.02, 0.5, 'Arb. Units',
                 va='center', rotation='vertical')
    else:
        fig.text(0.02, 0.5, ('{0} ({1})'.format(make_up(var), Iunit)),
                 va='center', rotation='vertical')

    # Set limits of 'diff' window
    for i, method in enumerate(methods):
        if method == 'diff':
            # symmetrize error scale:
            # auto-zoom on min, max, but discard first and last centile (case of spikes / divergences)
            Idiff = Idiffs[i]
            if discard_centile:
                Idiff_sorted = np.sort(Idiff[~np.isnan(Idiff)])
                ymax = max(abs(Idiff_sorted[-int(discard_centile*len(Idiff_sorted)//100)]), 
                           abs(Idiff_sorted[int(discard_centile*len(Idiff_sorted)//100)]))
            else:
                ymax = np.nanmax(abs(Idiff))
            ax1[i].set_ylim(-ymax*diff_scale_multiplier, ymax*diff_scale_multiplier)
        elif method == 'distance':
            if discard_centile:
                raise NotImplementedError('discard_centile not implemented for method=distance')
            _, ymax = ax1[i].get_ylim()
            ax1[i].set_ylim(0, ymax*diff_scale_multiplier)
        elif method == 'ratio':
            # auto-zoom on min, max, but discard first and last centile (case of spikes / divergences)
            Idiff = Idiffs[i]
            if discard_centile:
                Idiff_sorted = np.sort(Idiff[~np.isnan(Idiff)])
                ymin = Idiff_sorted[int(discard_centile*len(Idiff_sorted)//100)]
                ymax = Idiff_sorted[-int(discard_centile*len(Idiff_sorted)//100)]
            else:
                ymin = np.nanmin(Idiff)
                ymax = np.nanmax(Idiff)
            ax1[i].set_ylim(bottom=((ymin-1)*diff_scale_multiplier+1), 
                            top=((ymax-1)*diff_scale_multiplier+1))
#            ymax = max(abs(Idiff_sorted[len(Idiff_sorted)//100]-1),
#                       abs(Idiff_sorted[len(-Idiff_sorted)//100]-1))
#            ax1[i].set_ylim(ymax*diff_scale_multiplier+1, -ymax*diff_scale_multiplier+1)

    if title:
        fig.suptitle(title)

    # Fix format
    fix_style('origin', ax=ax0)
    for ax1i in ax1:
        fix_style('origin', ax=ax1i)
    plt.tight_layout()
    if title:
        plt.subplots_adjust(left=0.15, top=0.92)
    else:
        plt.subplots_adjust(left=0.15)

    # Plot difference text     
    for i, method in enumerate(methods):
        if method == 'diff':
            difftext = 'diff'
            ax1[i].axhline(y=0, color='grey', zorder=-1)
        elif method == 'distance':
            difftext = 'distance'
            ax1[i].axhline(y=0, color='grey', zorder=-1)
        elif method == 'ratio':
            difftext = 'ratio'
            ax1[i].axhline(y=1, color='grey', zorder=-1)
                    
        # Show residualget_residual
        if show_residual:
            difftext += ' (residual={0:.2g})'.format(get_residual(s1, s2, var=var, norm='L2',
                          ignore_nan=True, diff_window=diff_window))
        pos = ax1[i].get_position()
        fig.text(0.09, pos.ymax+0.02, difftext)

    # Add cursors
    axes = [ax0] + ax1
    fig.cursors = MultiCursor(fig.canvas, axes,
                              color='r', lw=1*lw_multiplier, alpha=0.2, horizOn=False,
                              vertOn=True)
    if show:
        plt.show()
    if save:
        fig.savefig(save)
        if not show: 
            plt.close(fig) #to avoid memory load if 

    # Return graphs
    return fig, axes


def averageDistance(s1, s2, var='radiance'):
    ''' Return the average distance between two spectra.
    It's important to note for fitting that if averageDistance(s1, s2)==0 
    then s1 = s2

    Parameters    
    ----------

    s1, s2: Spectrum objects
        spectra to be compared

    var: str, optional
        spectral quantity (ex: 'radiance', 'transmittance'...)

    Returns    
    -------

    distance: float
        Average distance as in the following equation:

        .. math::

            dist = \\frac{\\sqrt{\\sum_i {(s1_i-s2_i)^2}}}{N}.

    '''
    
    warn(FutureWarning('will be deprecated in future versions. Use get_residual(s1, s2, norm=L2) instead'))
    
    distArray = get_diff(s1, s2, var)
    distArray_Y2 = distArray[1]*distArray[1]
    N = np.size(distArray[1])
    distance = np.sqrt(np.sum(distArray_Y2))/N
    return distance



# %% Function to compare Spectra including conditions and lines


def compare_spectra(first, other, spectra_only=False, plot=True, wunit='default',
                 verbose=True, rtol=1e-5, ignore_nan=False, ignore_outliers=False,
                 normalize=False, **kwargs):
    ''' Compare Spectrum with another Spectrum object

    Parameters    
    ----------

    first: type Spectrum
        a Spectrum to be compared

    other: type Spectrum
        another Spectrum to compare with

    spectra_only: boolean, or str
        if ``True``, only compares spectral quantities (in the same waveunit)
        and not lines or conditions. If str, compare a particular quantity
        name. If False, compare everything (including lines and conditions
        and populations). Default ``False``

    plot: boolean
        if ``True``, use plot_diff to plot all quantities for the 2 spectra
        and the difference between them. Default ``False``.

    wunit: 'nm', 'cm-1', 'default'
        in which wavespace to compare (and plot). If default, natural wavespace
        of first Spectrum is taken

    rtol: float
        relative difference to use for spectral quantities comparison

    ignore_nan: boolean
        if ``True``, nans are ignored when comparing spectral quantities

    ignore_outliers: boolean, or float
        if not False, outliers are discarded. i.e, output is determined by::

            out = (~np.isclose(I, Ie, rtol=rtol, atol=0)).sum()/len(I) < ignore_outliers

    normalize: bool
        Normalize the spectra to be ploted 

    Other Parameters
    ----------------

    kwargs: dict
        arguments are forwarded to :func:`~radis.spectrum.compare.plot_diff`

    Returns
    -------

    equals: boolean
        return True if spectra are equal (respective to tolerance defined by 
        rtol and other input conditions)


    Examples
    --------

    Compare two Spectrum objects, or specifically the transmittance::

        s1.compare_with(s2)
        s1.compare_with(s2, 'transmittance')


    Note that you can also simply use `s1 == s2`, that uses 
    :meth:`~radis.spectrum.spectrum.Spectrum.compare_with` internally::

        s1 == s2       # will return True or False

    '''
    

    # Check inputs
    if not 0 <= ignore_outliers < 1:
        raise ValueError('ignore_outliers should be < 1, or False')
    if not is_spectrum(other):
        raise TypeError('2nd object is not a Spectrum: got class {0}'.format(
            other.__class__))
    if isinstance(spectra_only, string_types):   # case where we compare all quantities
        if not spectra_only in first.get_vars():
            raise ValueError('{0} is not a spectral quantity in our Spectrum ({1})'.format(
                spectra_only, first.get_vars()))
        if not spectra_only in other.get_vars():
            raise ValueError('{0} is not a spectral quantity in the other Spectrum ({1})'.format(
                spectra_only, other.get_vars()))
    if verbose:  # print conditions
        what = spectra_only if isinstance(
            spectra_only, string_types) else 'all quantities'
        msg = 'compare {0} with rtol={1}'.format(what, rtol)
        if ignore_nan:
            msg += ', ignore_nan'
        if ignore_outliers:
            msg += ', ignore_outliers={0}'.format(ignore_outliers)
        print(msg)
    if not plot and len(kwargs) > 0:
        raise ValueError('Unexpected argument: {0}'.format(kwargs))

    if wunit == 'default':
        wunit = first.get_waveunit()

    def _compare_dataframes(df1, df2, name):
        ''' 

        Parameters    
        ----------

        df1, df2: pandas Dataframe
            lines, or vib/rovib levels dataframes 

        name: str
            for error message
        '''

#            if compare_lists(df1.keys(), df2.keys(), verbose=False) != 1:
#                if verbose: print('... keys in {0} dont match:'.format(name))
#                compare_lists(list(df1.keys()), list(df2.keys()),
#                              verbose=True)
#                out = False
#            elif compare_lists(df1.index, df2.index, verbose=False) != 1:
#                if verbose: print('... index in {0} dont match:'.format(name))
#                compare_lists(list(df1.index), list(df2.index),
#                              verbose=True)
#                out = False
#            else:
#                out = (df1 == df2).all().all()
#
#            return out

        from pandas.util.testing import assert_frame_equal

        try:
            assert_frame_equal(df1.sort_index(axis=0).sort_index(axis=1),
                               df2.sort_index(axis=0).sort_index(axis=1),
                               check_names=True, 
                               check_column_type=False,  # solves problem in Python 2/3 dataframes (unicode/str)
                               )
            out = True

        except AssertionError as err:
            if verbose:
                print('Comparing ', name)
                print(err.args[0])
            out = False

        return out

    def _compare_variables(I, Ie):
        ''' Compare spectral quantities I and Ie '''

        if ignore_nan:
            b = ~(np.isnan(I) + np.isnan(Ie))
            I = I[b]
            Ie = Ie[b]

        if ignore_outliers:
            out = (~np.isclose(I, Ie, rtol=rtol, atol=0)).sum() / \
                len(I) < ignore_outliers
        else:
            out = np.allclose(I, Ie, rtol=rtol, atol=0)

        return bool(out)

    b = True
    if isinstance(spectra_only, string_types):  # compare this quantity
        vars = [spectra_only]
    else:                    # compare all quantities
        b = set(first.get_vars()) == set(other.get_vars())
        if not b and verbose:
            print('... list of quantities dont match: {0} vs {1}'.format(
                first.get_vars(), other.get_vars()))
        vars = [k for k in first.get_vars() if k in other.get_vars()]

    if spectra_only:
        # Compare spectral variables
        # -----------
        for k in vars:
            w, q = first.get(k, wunit=wunit)
            w0, q0 = other.get(k, wunit=wunit)
            if len(w) != len(w0):
                print(
                    'Wavespaces have different length (for {0}: {1} vs {2})'.format(
                            k, len(w), len(w0)))
                b1 = False
            else:
                b1 = np.allclose(w, w0, rtol=rtol, atol=0)
                b1 *= _compare_variables(q, q0)
                if not b1 and verbose:
                    error = np.nanmax(abs(q/q0-1))
                    avgerr = np.nanmean(abs(q/q0-1))
                    print('...', k, 'dont match (up to {0:.3}% diff.,'.format(
                        error*100)+' average {0:.3f}%)'.format(avgerr*100))
            b *= b1

            if plot:
                try:
                    plot_diff(first, other, var=k, wunit=wunit,
                              normalize=normalize, verbose=verbose,
                              **kwargs)
                except:
                    print('... couldnt plot {0}'.format(k))

    else:
        # Compare spectral variables
        # -----------
        for k in vars:
            w, q = first.get(k, wunit=wunit)
            w0, q0 = other.get(k, wunit=wunit)
            if len(w) != len(w0):
                print(
                    'Wavespaces have different length (for {0}: {1} vs {2})'.format(
                            k, len(w), len(w0)))
                b1 = False
            else:
                b1 = np.allclose(w, w0, rtol=rtol, atol=0)
                b1 *= _compare_variables(q, q0)
                if not b1 and verbose:
                    error = np.nanmax(abs(q/q0-1))
                    print('...', k, 'dont match (up to {0:.3f}% diff.)'.format(
                        error*100))
            b *= b1

            if plot:
                try:
                    plot_diff(first, other, var=k, wunit=wunit,
                              normalize=normalize, verbose=verbose,
                              **kwargs)
                except:
                    print('... there was an error while plotting {0}'.format(k))

        # Compare conditions and units
        # -----------
        verbose_dict = 'if_different' if verbose else False
        b1 = (compare_dict(first.conditions, other.conditions, verbose=verbose_dict) == 1)
        b2 = (compare_dict(first.cond_units, other.cond_units, verbose=verbose_dict) == 1)
        b3 = (compare_dict(first.units, other.units, verbose=verbose_dict) == 1)
        if not b1 and verbose:
            print('... conditions dont match')
        if not b2 and verbose:
            print('... conditions units dont match')
        if not b3 and verbose:
            print('... units dont match')
        b *= b1 * b2 * b3

        # Compare lines
        # -----------
        if first.lines is None and other.lines is None:
            b4 = True
        elif first.lines is None:
            b4 = False
        elif other.lines is None:
            b4 = False
        else:
            b4 = _compare_dataframes(first.lines, other.lines, 'lines')
        if not b4 and verbose:
            print('... lines dont match')
        b *= b4

        # Compare populations
        # -----------
        if first.populations is None and other.populations is None:
            b5 = True
        elif first.populations is None:
            b5 = False
        elif other.populations is None:
            b5 = False
        else:
            # Compare keys in populations
            b5 = True
            if compare_lists(first.populations,
                             other.populations,
                             verbose='if_different') == 1:
                # same molecules, compare isotopes
                for molecule, isotopes in first.populations.items():
                    if compare_lists(isotopes,
                                     other.populations[molecule],
                                     verbose='if_different') == 1:
                        # same isotopes, compare electronic states
                        for isotope, states in isotopes.items():
                            if compare_lists(states,
                                             other.populations[molecule][isotope],
                                             verbose='if_different') == 1:
                                # same electronic states, compare levels + other information
                                for state, content in states.items():
                                    for k, v in content.items():
                                        if k in ['vib', 'rovib']:
                                            b5 *= _compare_dataframes(
                                                v,
                                                other.populations[molecule][isotope][state][k],
                                                'populations of {0}({1})(iso{2})'.format(
                                                    molecule, state, isotope))
                                        else:
                                            b5 *= (
                                                v == other.populations[molecule][isotope][state][k])
                            else:
                                b5 = False
                                if verbose:
                                    print('{0}(iso{1}) states are different (see above)'.format(
                                        molecule, isotope))
                    else:
                        b5 = False
                        if verbose:
                            print(
                                '{0} isotopes are different (see above)'.format(molecule))
            else:
                b5 = False
                if verbose:
                    print('Molecules are different (see above)')
        if not b5 and verbose:
            print('... populations dont match (see detail above)')
        b *= b5

        # Compare slit
        # -----------

        if len(first._slit) == len(other._slit) == 0:
            # no slit anywhere
            b6 = True
        elif len(first._slit) != len(other._slit):
            b6 = False
            if verbose:
                print('A spectrum has slit function array but the other doesnt')
        else:
            ws, Is = first.get_slit()
            ws0, Is0 = other.get_slit()
            if len(ws) != len(ws0):
                if verbose:
                    print('Slits have different length')
                b6 = False
            else:
                b6 = np.allclose(ws, ws0, rtol=rtol, atol=0)
                b6 *= _compare_variables(Is, Is0)
                if not b6 and verbose:
                    print('Slit functions dont match')
        b *= b6

    return bool(b)







if __name__ == '__main__':

    from radis.test.spectrum.test_compare import _run_testcases
    print(('Test compare.py: ', _run_testcases(verbose=True)))



# -*- coding: utf-8 -*-
"""
Summary
-------

Public (front-end) functions to calculate Spectrum with HITRAN / CDSD databanks.
Uses the SpectrumFactory classe from `factory.py`, Spectrum from `spectrum.py`
and line survey from `line_survey.py` 

Routine Listing
---------------

:func:`~radis.lbl.calc.calc_spectrum`

-------------------------------------------------------------------------------

"""

from __future__ import print_function, absolute_import, division, unicode_literals

from radis.lbl.factory import SpectrumFactory
from radis.spectrum import Spectrum, is_spectrum
from radis.phys.blackbody import planck, planck_wn
from radis.phys.convert import nm2cm
from radis.misc.utils import DatabankNotFound
import numpy as np
from numpy import exp, arange, allclose, abs, diff, zeros_like, ones_like
from warnings import warn

# %%


def calc_spectrum(wavenum_min=None,
                  wavenum_max=None,
                  wavelength_min=None,
                  wavelength_max=None,
                  Tgas=None,
                  Tvib=None,
                  Trot=None,
                  pressure=1.01325,
                  overpopulation=None,
                  molecule='',
                  isotope='all',
                  mole_fraction=1,
                  path_length=1,
                  databank='fetch',
                  slit=None,
                  plot=None,
                  medium='air',
                  wstep=0.01,
                  name=None,
                  use_cached=True,
                  **kwargs):
    ''' Multipurpose function to calculate :class:`~radis.spectrum.spectrum.Spectrum`
    under equilibrium, or non-equilibrium, with or without overpopulation. 
    It's a wrapper to :class:`~radis.lbl.factory.SpectrumFactory` class. 
    For advanced used, please refer to the aforementionned class. 

    Parameters
    ----------

    wavenum_min: float (cm-1)
        minimum wavenumber to be processed in cm^-1
    wavenum_max: float (cm-1)
        maximum wavenumber to be processed in cm^-1

    wavelength_min: float (nm)
        minimum wavelength to be processed in nm. Wavelength if in 'air' or 
        'vacuum' depending of the value of the parameter 'medium='

    wavelength_max: float (nm)
        maximum wavelength to be processed in nm. Wavelength if in 'air' or 
        'vacuum' depending of the value of the parameter 'medium='

    Tgas: float (K)
        Gas temperature. If non equilibrium, is used for Ttranslational. 
        Default 300K

    Tvib: float (K)
        Vibrational temperature. If None, equilibrium calculation is run with Tgas

    Trot: float (K)
        Rotational temperature. If None, equilibrium calculation is run with Tgas

    pressure: (bar)
        partial pressure of gas in bar. Default 1.01325 (1 atm)

    overpopulation: dict
        dictionary of overpopulation compared to the given vibrational temperature. 
        Ex with CO2::

        >>> overpopulation = {
        >>>    '(00`0`0)->(00`0`1)': 2.5,
        >>>     '(00`0`1)->(00`0`2)': 1,
        >>>     '(01`1`0)->(01`1`1)': 1,
        >>>     '(01`1`1)->(01`1`2)': 1,
        >>>     }

    molecule: int, str, or ``None``
        molecule id (HITRAN format) or name. If None, the molecule is infered
        from the database files being loaded. It is just good practice to fill
        it. Default ``None``.

    isotope: int, list, str of the form '1,2', or 'all'
        isotope id (sorted by relative density: (eg: 1: CO2-626, 2: CO2-636 for CO2).
        See HITRAN documentation for isotope list for all species. If 'all',
        all isotopes in database are used (this may result in larger computation
        times!). Default 'all'

    mole_fraction: float
        database species mole fraction. Default 1

    path_length: float (cm)
        slab size. Default 1.

    databank: str
        can be either: 

        - the name of a spectral database to use (must be registered 
          in your ``~/.radis``). See :ref:`Configuration file <label_lbl_config_file>`.

        - ``'fetch'``, to fetch automatically from [HITRAN-2016]_ through astroquery. 
        
        .. warning::
            
            [HITRAN-2016]_ is validd for low temperatures (typically < 700 K). For higher
            temperatures you may need [HITEMP-2010]_ 

        Default ``'fetch'``. See :class:`~radis.lbl.loader.DatabankLoader` for more 
        information on line databases, and :data:`~radis.misc.config.DBFORMAT` for 
        your ``~/.radis`` file format 

    slit: float, str, or ``None``
        if float, FWHM of a triangular slit function. If str, path to an 
        experimenta slit function. If None, no slit is applied. Default ``None``.

    plot: str
        any parameter such as 'radiance' (if slit is given), 'radiance_noslit', 
        'absorbance', etc...   Default ``None``

    medium: 'air', vacuum'
        propagating medium: choose whether to return wavelength in air or vacuum.
        Note that the  input wavelength range ("wavenum_min", "wavenum_max") changes. 
        Default 'air'

    Other Parameters
    ----------------

    wstep: cm-1
        Spacing of calculated spectrum. Default 0.01 cm-1

    name: str
        name of the case. If None, a unique ID is generated. Default ``None``

    use_cached: boolean
        use cached files for line database and energy database. Default ``True``

    **kwargs: other inputs forwarded to SpectrumFactory
        See :class:`~radis.lbl.factory.SpectrumFactory` documentation for more details on input

    Example:

    pseudo_continuum_threshold: float
        if not 0, first calculate a rough approximation of the spectrum, then
        moves all lines whose linestrength intensity is less than this threshold
        of the maximum in a semi-continuum. Values above 0.01 can yield significant
        errors, mostly in highly populated areas. 80% of the lines can typically
        be moved in a continuum, resulting in 5 times faster spectra. If 0,
        no semi-continuum is used. Default 0.

    Returns
    -------

    Returns a :class:`~radis.spectrum.spectrum.Spectrum` object::
    
    Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something 
    among ['radiance', 'radiance_noslit', 'absorbance', etc...]

    Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method 
    to plot it

    See [1]_ to get an overview of all Spectrum methods

    Notes
    -----

    Calculate line strenghts correcting the CDSD reference one. Then call
    the main routine that sums over all lines


    References
    ----------

    .. [1] RADIS doc: `Spectrum how to? <http://radis.readthedocs.io/en/latest/#the-spectrum-class>`__


    Examples
    --------

    Calculate a CO spectrum from the HITRAN database ::

        s = calc_spectrum(1900, 2300,         # cm-1
                          molecule='CO',
                          isotope='1,2,3',
                          pressure=1.01325,   # bar
                          Tgas=1000, 
                          mole_fraction=0.1, 
                          )
        s.apply_slit(0.5, 'nm')
        s.plot('radiance')
        
    This example uses the :py:meth:`~radis.spectrum.spectrum.Spectrum.apply_slit` 
    and :py:meth:`~radis.spectrum.spectrum.Spectrum.plot` methods. See also
    :py:meth:`~radis.spectrum.spectrum.Spectrum.line_survey`:: 
        
        s.line_survey(overlay='radiance')

    Refer to the online :ref:`Examples <label_examples>` for more cases. 

    '''

    # Check inputs

    # ... wavelengths / wavenumbers
    if ((wavelength_min is not None or wavelength_max is not None) and
            (wavenum_min is not None or wavenum_max is not None)):
        raise ValueError("Wavenumber and Wavelength both given... it's time to choose!")

    if (wavenum_min is None and wavenum_max is None):
        assert(wavelength_max is not None)
        assert(wavelength_min is not None)
        wavenum_min = nm2cm(wavelength_max)
        wavenum_max = nm2cm(wavelength_min)
    else:
        assert(wavenum_min is not None)
        assert(wavenum_max is not None)

    # ... temperatures

    if Tgas is None and Trot is None:
        raise ValueError(
            'Choose either Tgas (equilibrium) or Tvib / Trot (non equilibrium)')

    if Tvib is None and Trot is not None or Tvib is not None and Trot is None:
        raise ValueError('Choose both Tvib and Trot')
        
    if databank == 'fetch' and ((Tgas is not None and Tgas > 700) or 
                                (Tvib is not None and Tvib > 700) or 
                                (Trot is not None and Trot > 700)):
        warn("HITRAN is valid for low temperatures (typically < 700 K). "+\
             "For higher temperatures you may need HITEMP or CDSD. See the "+\
             "'databank=' parameter")
        
    # ... others
    if databank is None:
        raise ValueError('Give a databank name')

    if not 'save_memory' in kwargs:
        # no need to save intermediary results as 
         # factory is used once only
        kwargs['save_memory'] = True
       

    def _is_at_equilibrium():
        try:
            assert Tvib is None or Tvib == Tgas
            assert Trot is None or Trot == Tgas
            assert overpopulation is None
            if 'self_absorption' in kwargs:
                assert kwargs['self_absorption']  # == True
            return True
        except AssertionError:
            return False

    _equilibrium = _is_at_equilibrium()

    # Run calculations
    sf = SpectrumFactory(wavenum_min, wavenum_max, medium=medium,
                         molecule=molecule, isotope=isotope,
                         pressure=pressure,
                         wstep=wstep,
                         db_use_cached=use_cached,
                         **kwargs)
    if databank == 'fetch':       # mode to get databank without relying on  Line databases
        if _equilibrium:
            # Get line database from HITRAN
            # and partition functions from HAPI
            sf.fetch_databank(source='astroquery', format='hitran',
                              parfuncfmt='hapi',   # use HAPI partition functions for equilibrium
                              levelsfmt=None,      # no need to load energies
                              )
        else:
            # Also get line database from HITRAN, and calculate partition functions
            # with energy levels from built-in constants (not all molecules
            # are supported!)
            sf.fetch_databank(source='astroquery', format='hitran',
                              parfuncfmt='hapi',   # use HAPI partition functions for equilibrium
                              levelsfmt='radis',     # built-in spectroscopic constants
                              )
    else:   # manual mode: get from files defined in .radis
        sf.load_databank(databank, 
                         load_energies=not _equilibrium   # no need to load/calculate energies at eq.
                         )

    # Use the standard eq_spectrum / non_eq_spectrum functions
    if _equilibrium:
        s = sf.eq_spectrum(Tgas=Tgas, mole_fraction=mole_fraction, path_length=path_length,
                           name=name)
    else:
        s = sf.non_eq_spectrum(Tvib=Tvib, Trot=Trot, Ttrans=Tgas,
                               overpopulation=overpopulation,
                               mole_fraction=mole_fraction,
                               path_length=path_length,
                               name=name)

    if slit is not None:
        s.apply_slit(slit)
    if plot is not None and plot is not False:
        s.plot(plot)
    return s


# --------------------------
if __name__ == '__main__':

    from radis.test.lbl.test_calc import _run_testcases
    print(_run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Mon May  7 17:34:52 2018

@author: erwan
"""

from __future__ import absolute_import, unicode_literals, division, print_function
from radis.lbl import SpectrumFactory
from radis.misc.utils import DatabankNotFound
from radis.misc.printer import printm
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
import matplotlib.pyplot as plt
from os.path import exists
from shutil import rmtree

def test_retrieve_from_database(plot=True, verbose=True, warnings=True, *args, **kwargs):
    ''' Test autoretrieve from a database:
        
        first generate an empty :py:class:`~radis.tools.database.SpecDatabase`
        associated to a :py:class`~radis.lbl.factory.SpectrumFactory`, 
        then calculate a first spectrum, then calculate it again and make sure
        it is retrieved from the database
        '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()
        
    temp_database_name = 'temp_spec_database'

    try:
        if verbose:
            printm('>>> test_retrieve_from_database')

        assert not exists(temp_database_name)

        setup_test_line_databases()  # add HITEMP-CO2-TEST in ~/.radis if not there

        sf = SpectrumFactory(2284.2, 2284.6,
                             wstep=0.001,                # cm-1
                             pressure=20*1e-3,           # bar
                             cutoff=0,
                             path_length=0.1, 
                             mole_fraction=400e-6,
                             molecule='CO2',
                             isotope=[1], db_use_cached=True,
                             medium='vacuum',
                             broadening_max_width=10,
                             export_populations='rovib',
                             verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.init_databank('HITEMP-CO2-TEST')    # unlike load_databank, will automatically be built when needed
        db = sf.init_database(temp_database_name, autoretrieve=True)
        
        assert len(db) == 0
        
        # Calculate a first spectrum
        s1 = sf.non_eq_spectrum(2000, 1000)
        
        assert len(db) == 1
        # Note that the Spectrum in database is not s1 because populations
        # are not stored by default 
    
    
        # hack: now change the `autoretrieve` status to `force` to make sure 
        # the spectrum is retrieved. Else, an error will be raised    
        sf.autoretrievedatabase = 'force'
        
        # Calculate spectrum under the same conditions
        s2 = sf.non_eq_spectrum(2000, 1000)
        
        # Note that s1 == s2 won't work because populations are not stored
        # by default in the database
        assert s1.compare_with(s2, spectra_only=True)
        
        return True

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)
        
    finally:
        rmtree(temp_database_name)


def _run_testcases(verbose=True, plot=True):
    
    test_retrieve_from_database(plot=plot, verbose=verbose)

if __name__ == '__main__':
    _run_testcases()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Handy functions for manipulation of :class:`~radis.spectrum.spectrum.Spectrum` objects

Routine Listing
---------------

Operators:

- :py:func:`~radis.spectrum.operations.multiply` : simply use ``s*2``
- :py:func:`~radis.spectrum.operations.add_constant` : simply use ``s+1``
- :py:func:`~radis.spectrum.operations.add_array` : simply use ``s+a``
- :py:func:`~radis.spectrum.operations.add_spectra` : simply use ``s1+s2`` 
- :py:func:`~radis.spectrum.operations.substract_spectra` : simply use ``s1-s2``

Note that these operators are purely algebraic and should not be used in place
of the line-of-sight functions, i.e, :py:func:`~radis.los.slabs.SerialSlabs` (``>``)
and :py:func:`~radis.los.slabs.MergeSlabs` (``//``)

Functions to manipulate one spectrum:
    
- :py:func:`~radis.spectrum.operations.crop`
- :py:func:`~radis.spectrum.operations.offset`

Play with baselines:
    
- :py:func:`~radis.spectrum.operations.get_baseline`
- :py:func:`~radis.spectrum.operations.sub_baseline`
    
Functions to discard all but one spectral quantity:
    
- :py:func:`~radis.spectrum.operations.Transmittance`
- :py:func:`~radis.spectrum.operations.Transmittance_noslit`
- :py:func:`~radis.spectrum.operations.Radiance`
- :py:func:`~radis.spectrum.operations.Radiance_noslit`

Keeps all spectral quantities, but make emission equal to 0 (useful when
calculating contributions of line of sight slabs):
    
- :py:func:`~radis.spectrum.operations.PerfectAbsorber`


Examples
--------

Most of these functions are implemented with the standard operators. Ex::
    
    ((s_exp - 0.1)*10).plot()   # works for a Spectrum s_exp

-------------------------------------------------------------------------------


"""

from __future__ import print_function, absolute_import, division, unicode_literals
#from radis.misc.curve import curve_substract, curve_add
from radis.spectrum import Spectrum
from radis.phys.convert import (cm2nm, nm2cm, cm2nm_air, nm_air2cm, air2vacuum, vacuum2air,
                                dcm2dnm, dnm2dcm)
from numpy import ones_like, hstack
from warnings import warn


# %% Filter Spectra

def Transmittance(s):
    # type: (Spectrum) -> Spectrum
    ''' Returns a new Spectrum with only the ``transmittance`` component of ``s`` 
    
    Parameters
    ----------
    
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object
        
    Returns
    -------
    
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only the ``transmittance``,
        ``absorbance`` and/or ``abscoeff`` part of ``s``, where ``radiance_noslit`` ,
        ``emisscoeff`` and ``emissivity_noslit`` (if they exist) have been set to 0
    
    See Also
    --------
    
    :py:func:`~radis.spectrum.operations.Transmittance_noslit`

    '''

    return s.copy(copy_lines=True, quantity='transmittance')

def Transmittance_noslit(s):
    ''' Returns a new Spectrum with only the ``transmittance_noslit`` component of ``s`` 
    
    Parameters
    ----------
    
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object
        
    Returns
    -------
    
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only 
        ``transmittance_noslit`` defined
 
    See Also
    --------
    
    :py:func:`~radis.spectrum.operations.Transmittance`
    '''

    return s.copy(copy_lines=True, quantity='transmittance_noslit')

def Radiance(s):
    ''' Returns a new Spectrum with only the ``radiance`` component of ``s`` 
    
    Parameters
    ----------
    
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object
        
    Returns
    -------
    
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only ``radiance``
        defined
 
    See Also
    --------
    
    :py:func:`~radis.spectrum.operations.Radiance_noslit`
    '''

    return s.copy(copy_lines=True, quantity='radiance')

def Radiance_noslit(s):
    ''' Returns a new Spectrum with only the ``radiance_noslit`` component of ``s`` 
    
    Parameters
    ----------
    
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object
        
    Returns
    -------
    
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only ``radiance_noslit``
        defined
        
    See Also
    --------
    
    :py:func:`~radis.spectrum.operations.Radiance`

    '''

    return s.copy(copy_lines=True, quantity='radiance_noslit')

# Useful for determining line-of-sight contributions:

def PerfectAbsorber(s):
    ''' Makes a new Spectrum with the same transmittance/absorbance as Spectrum
    ``s``, but with radiance set to 0. 
    Useful to get contribution of different slabs in line-of-sight 
    calculations (see example).
    
    .. note:
        
        formerly named "Transmittance", but "Transmittance(s)" wouldnt 
        return the Transmittance exactly
    
    Parameters
    ----------
    
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object
        
    Returns
    -------
    
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only the ``transmittance``,
        ``absorbance`` and/or ``abscoeff`` part of ``s``, where ``radiance_noslit`` ,
        ``emisscoeff`` and ``emissivity_noslit`` (if they exist) have been set to 0
        
    Examples
    --------
    
    Let's say you have a total line of sight::
    
        s_los = s1 > s2 > s3     
        
    If you now want to get the contribution of ``s2`` to the line-of-sight emission,
    you can do::
        
        (s2 > PerfectAbsorber(s3)).plot('radiance_noslit')
        
    And the contribution of ``s1`` would be::
        
        (s1 > PerfectAbsorber(s2>s3)).plot('radiance_noslit') 
    
    See more examples in :ref:`Line-of-Sight module <label_los_index>`
    
    '''
    
    s_tr = s.copy()
    for k in ['radiance_noslit', 'emisscoeff', 'emissivity_noslit']:
        if k in s_tr._q:
            s_tr._q[k] *= 0
            
    for k in ['radiance', 'emissivity']:
        if k in s_tr._q_conv:
            s_tr._q_conv[k] *= 0

    # Deactivate equilibrium conditions (so Kirchoff's law cannot be used anymore)
    s_tr.conditions['thermal_equilibrium'] = False
    
    s_tr.name = 'PureAbsorber({0})'.format(s.get_name())
    
    return s_tr
    

# %% Change wavelength
    

def crop(s, wmin=None, wmax=None, wunit=None, medium=None,
         inplace=False):
    # type: (Spectrum, float, float, str, str, bool) -> Spectrum
    ''' Crop spectrum to ``wmin-wmax`` range in ``wunit``
    
    Parameters
    ----------
    
    s: Spectrum object
        object to crop
    
    wmin, wmax: float, or None
        boundaries of spectral range (in ``wunit``)
        
    wunit: ``'nm'``, ``'cm-1'``
        which waveunit to use for ``wmin, wmax``. Default ``default``: 
        just use the Spectrum wavespace. 

    medium: 'air', vacuum'
        necessary if cropping in 'nm'
        
    Other Parameters
    ----------------
    
    inplace: bool
        if ``True``, modifiy ``s`` directly. Else, returns a copy.
    
    Returns
    -------
    
    s_crop: Spectrum
        a cropped Spectrum.
        if using ``inplace``, then ``s_crop`` and ``s`` are still the same object
    
    Examples
    --------
    
    ::
        
        crop(s, 420, 480, 'nm', 'air')
        
    Or in ``cm-1``::
        
        crop(s, 2000, 2300, 'cm-1')
    
    '''
    
    # Check inputs
    if wmin is None and wmax is None:
        raise ValueError('Choose at least `wmin=` or `wmax=`')
    if wunit is None:
        raise ValueError('Please precise unit for wmin and wmax with `unit=`')
    assert wunit in ['nm', 'cm-1']
    if wunit == 'nm' and medium is None:
        raise ValueError("Precise wavelength medium with medium='air' or "+\
                         "medium='vacuum'")
    if (wmin is not None and wmax is not None) and wmin >= wmax:
        raise ValueError('wmin should be < wmax (Got: {0:.2f}, {1:.2f})'.format(
                wmin, wmax))
    
    if len(s._q)>0 and len(s._q_conv)>0:
        raise NotImplementedError('Cant crop this Spectrum as there are both convoluted '+\
                                  'and not convoluted quantities stored')
        # Could bring unexpected errors... For instance, if cropping both 
        # with slit and without slit  quantities to the same waverange, 
        # reapplying the slit in 'valid' mode would reduce the wavelength range 
        # of the convoluted quantities 
        # Implementation: better ask User to drop some of the quantities themselves
    
    if not inplace:
        s = s.copy()
    
    # Convert wmin, wmax to Spectrum wavespace    
    # (deal with cases where wavelength are given in 'air' or 'vacuum')
    # TODO @dev: rewrite with wunit='cm-1', 'nm_air', 'nm_vac'
    waveunit = s.get_waveunit()
    wmin0, wmax0 = wmin, wmax
    if wunit == 'nm' and waveunit == 'cm-1':
        if medium == 'air':
            if wmin: wmin = nm_air2cm(wmax0)   # reverted
            if wmax: wmax = nm_air2cm(wmin0)   # reverted
        else:
            if wmin: wmin = nm2cm(wmax0)   # reverted
            if wmax: wmax = nm2cm(wmin0)   # reverted
    elif wunit == 'cm-1' and waveunit == 'nm':
        if s.get_medium() == 'air':
            if wmin: wmin = cm2nm_air(wmax0)   # nm in air
            if wmax: wmax = cm2nm_air(wmin0)   # nm in air
        else:
            if wmin: wmin = cm2nm(wmax0)   # get nm in vacuum
            if wmax: wmax = cm2nm(wmin0)   # get nm in vacuum
    elif wunit == 'nm' and waveunit == 'nm':
        if s.get_medium() == 'air' and medium == 'vacuum':
            # convert from given medium ('vacuum') to spectrum medium ('air')
            if wmin: wmin = vacuum2air(wmin0)
            if wmax: wmax = vacuum2air(wmax0)
        elif s.get_medium() == 'vacuum' and medium == 'air':
            # the other way around
            if wmin: wmin = air2vacuum(wmin0)
            if wmax: wmax = air2vacuum(wmax0)
    else:
        assert wunit == waveunit           # correct wmin, wmax
    
    # Crop non convoluted
    if len(s._q)>0:
        b = ones_like(s._q['wavespace'], dtype=bool)
        if wmin: b *= (wmin <= s._q['wavespace'])
        if wmax: b *= (s._q['wavespace'] <= wmax)
        for k, v in s._q.items():
            s._q[k] = v[b]
  
    # Crop convoluted
    if len(s._q_conv)>0:
        b = ones_like(s._q_conv['wavespace'], dtype=bool)
        if wmin: b *= (wmin <= s._q_conv['wavespace'])
        if wmax: b *= (s._q_conv['wavespace'] <= wmax)
        for k, v in s._q_conv.items():
            s._q_conv[k] = v[b]
            
    return s
    



# %% Algebric operations on Spectra
    
def _get_unique_var(s, var, inplace):
    ''' Returns the unique spectral quantity in the Spectrum ``s``. If there are more, 
    raises an error.
    '''
    # If var is undefined, get it if there is no ambiguity
    if var is None:
        quantities = s.get_vars()
        if len(quantities)>1:
            raise KeyError('There is an ambiguity with the Spectrum algebraic operation. '+\
                            'There should be only one var in Spectrum {0}. Got {1}\n'.format(
                                    s.get_name(), s.get_vars())+\
                            "Think about using 'Transmittance(s)', 'Radiance(s)', etc.")
        elif len(quantities)==0:
            raise KeyError('No spectral quantity defined in Spectrum {0}'.format(s.get_name()))
        else:
            var = quantities[0]
    # If inplace, assert there is only one var
    if inplace and len(s.get_vars())>1:
        raise ValueError('Cant modify inplace one spectral quantity of a Spectrum '+\
                         "that contains several ({0}). Use 'inplace=False'".format(
                                 s.get_vars()))
    return var

def multiply(s, coef, var=None, inplace=False):
    '''Multiply s[var] by the float 'coef'

    Parameters
    ----------
    s: Spectrum objects
        The spectra to multiply.
    coef: float
        Coefficient of the multiplication.
    var: str, or ``None``
        'radiance', 'transmittance', ... If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy.
        Default ``False``

    Returns
    -------
    s : Spectrum
        Spectrum object where intensity of s['var'] is multiplied by coef
        If ``inplace=True``, ``s`` has been modified directly.

    '''
    # Check input
    var = _get_unique_var(s, var, inplace)
        
    if not inplace:
        s = s.copy(quantity=var)
#        if name is not None:
#            s.name = name
    
    # Multiply inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), copy=False)  
    I *= coef         # @dev: updates the Spectrum directly because of copy=False

    return s

def add_constant(s, cst, unit=None, var=None, wunit='nm', inplace=False):
    '''Return a new spectrum with a constant added to s[var]. 
    Equivalent to::
        
        s + constant

    Parameters    
    ----------
    s: Spectrum objects
        Spectrum you want to modify
    cst: Float
        Constant to add.
    unit: str
        unit for ``cst``. If ``None``, uses the default unit in ``s`` for 
        variable ``var``.
    var: str, or ``None``
        'radiance', 'transmittance', ... If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    wunit: str
        'nm'or 'cm-1'
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy. 
        Default ``False``

    Returns    
    -------
    s : Spectrum
        Spectrum object where cst is added to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.

    Notes   
    -----
    
    Use only for rough work. If you want to work properly with spectrum 
    objects, see :py:meth:`~radis.los.slabs.MergeSlabs`.
    '''
    # Check input
    var = _get_unique_var(s, var, inplace)

    # Convert to Spectrum unit
    if unit is not None:
        Iunit = s.units[var]
        if unit != Iunit:
            from radis.phys.convert import conv2
            cst = conv2(cst, unit, Iunit)

    if not inplace:
        s = s.copy(quantity=var)
    
    # Add inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), copy=False)  
    I += cst
    # @dev: updates the Spectrum directly because of copy=False
    
#    s.name = '{0}+{1}'.format(s.get_name(), cst)

    return s

def add_array(s, a, unit=None, var=None, wunit='nm', inplace=False):
    '''Return a new spectrum with a constant added to s[var]. 
    Equivalent to::
        
        s + array

    Parameters    
    ----------
    s: Spectrum objects
        Spectrum you want to modify
    a: numpy array
        array to add. Must have the same length as variable ``var`` in Spectrum 
        ``s``
    unit: str
        unit for ``a``. If ``None``, uses the default unit in ``s`` for 
        variable ``var``.
    var: str, or ``None``
        'radiance', 'transmittance', ... If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    wunit: str
        'nm'or 'cm-1'
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy. 
        Default ``False``

    Returns    
    -------
    s : Spectrum
        Spectrum object where array ``a`` is added to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.

    Notes   
    -----
    Use only for rough work. If you want to work properly with spectrum 
    objects, see MergeSlabs.
    
    Examples
    --------
    
    Add Gaussian noise to your Spectrum (assuming there is only one spectral
    quantity defined)::
        
        s += np.random.normal(0,1,len(s.get_wavelength()))

    '''
    # Check input
    var = _get_unique_var(s, var, inplace)

    # Convert to Spectrum unit
    if unit is not None:
        Iunit = s.units[var]
        if unit != Iunit:
            from radis.phys.convert import conv2
            a = conv2(a, unit, Iunit)

    if not inplace:
        s = s.copy(quantity=var)
    
    # Add inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), copy=False)  
    I += a
    # @dev: updates the Spectrum directly because of copy=False

    return s

def sub_baseline(s, left, right, unit=None, var=None, wunit='nm',
                  inplace=False):
    '''Return a new spectrum with a baseline substracted to s[var] 
    
    Parameters    
    ----------
    
    s: Spectrum objects
        Spectrum you want to modify
    left: Float
        Constant to substract on the left of the spectrum.
    right: Float
        Constant to substract on the right of the spectrum.
    unit: str
        unit for ``cst``. If ``None``, uses the default unit in ``s`` for 
        variable ``var``.
    var: str
        'radiance', 'transmittance', ...  If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    wunit: str
        'nm'or 'cm-1'
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy. 
        Default ``False``
        
    Returns    
    -------
    
    s: Spectrum
        Spectrum object where the baseline was substracted to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.

    Notes    
    -----
    Use only for rough work. 
    
    See Also
    --------
    
    :py:func:`~radis.spectrum.operations.get_baseline`
    '''
    import numpy as np
    # Check input
    var = _get_unique_var(s, var, inplace)
        
    # Convert to Spectrum unit
    if unit is not None:
        Iunit = s.units[var]
        if unit != Iunit:
            from radis.phys.convert import conv2
            left = conv2(left, unit, Iunit)
            right = conv2(right, unit, Iunit)

    if not inplace:
        s = s.copy(quantity=var)
       
    # @EP: 

    # Substract inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), copy=False)  
    I -= np.linspace(left, right, num=np.size(I))
    # @dev: updates the Spectrum directly because of copy=False

    return s
    
def add_spectra(s1, s2, var=None, wunit='nm', force=False):
    '''Return a new spectrum with ``s2`` added to ``s1``. 
    Equivalent to::
        
        s1 + s2
    
    .. warning::
        we are just algebrically adding the quantities. If you want to merge
        spectra while preserving the radiative transfer equation, see 
        :func:`~radis.los.slabs.MergeSlabs` and :func:`~radis.los.slabs.SerialSlabs`
    
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
        Spectrum you want to substract
    var: str
        quantity to manipulate: 'radiance', 'transmittance', ... If ``None``, 
        get the unique spectral quantity of ``s1``, or the unique spectral
        quantity of ``s2``, or raises an error if there is any ambiguity
    wunit: str
        'nm'or 'cm-1'
        
    Returns    
    -------
    
    s: Spectrum
        Spectrum object with the same units and waveunits as ``s1``
        
    See Also
    --------
    
    :func:`~radis.los.slabs.MergeSlabs`,
    :func:`~radis.spectrum.operations.substract_spectra`
    
    '''
    
    # Get variable
    if var is None:
        try:
            var = _get_unique_var(s2, var, inplace=False)    # unique variable of 2nd spectrum  
        except KeyError:
            var = _get_unique_var(s1, var, inplace=False)    # if doesnt exist, unique variable of 1st spectrum
            # if it fails, let it fail
    # Make sure it is in both Spectra
    if var not in s1.get_vars():
        raise KeyError('Variable {0} not in Spectrum {1}'.format(var, s1.get_name()))
    if var not in s2.get_vars():
        raise KeyError('Variable {0} not in Spectrum {1}'.format(var, s1.get_name()))
        
    if var in ['transmittance_noslit', 'transmittance'] and not force:
        raise ValueError('It does not make much physical sense to sum transmittances. Are '+\
                         'you sure of what you are doing? See also `//` (MergeSlabs), `>` '+\
                         "(SerialSlabs) and `concat_spectra`. If you're sure, use `force=True`")
    
    # Use same units 
    Iunit = s1.units[var]            
    wunit = s1.get_waveunit()
    medium = s1.get_medium()

    # Resample
    if wunit == 'nm':
        s2 = s2.resample(s1.get_wavelength(), 'nm', inplace=False)
    elif wunit == 'cm-1':
        s2 = s2.resample(s1.get_wavenumber(), 'cm-1', inplace=False)
    else:
        raise ValueError('Unexpected wunit: {0}'.format(wunit))
    
    # Add
    w1, I1 = s1.get(var=var, Iunit=Iunit, wunit=wunit, medium=medium)
    w2, I2 = s2.get(var=var, Iunit=Iunit, wunit=wunit, medium=medium)

    name = s1.get_name()+'+'+s2.get_name()
    
    sub = Spectrum.from_array(w1, I1 + I2, var, 
                               waveunit=wunit, 
                               unit=Iunit,
                               conditions={'medium' : medium}, 
                               name=name)
#    warn("Conditions of the left spectrum were copied in the substraction.", Warning)
    return sub

def substract_spectra(s1, s2, var=None, wunit='nm'):
    '''Return a new spectrum with ``s2`` substracted from ``s1``. 
    Equivalent to::
        
        s1 - s2
    
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
        Spectrum you want to substract
    var: str
        quantity to manipulate: 'radiance', 'transmittance', ... If ``None``, 
        get the unique spectral quantity of ``s1``, or the unique spectral
        quantity of ``s2``, or raises an error if there is any ambiguity
    wunit: str
        'nm'or 'cm-1'
        
    Returns    
    -------
    
    s: Spectrum
        Spectrum object with the same units and waveunits as ``s1``
        
    See Also
    --------
    
    :func:`~radis.spectrum.operations.add_spectra`
    
    '''

    # Get variable
    if var is None:
        try:
            var = _get_unique_var(s2, var, inplace=False)    # unique variable of 2nd spectrum  
        except KeyError:
            var = _get_unique_var(s1, var, inplace=False)    # if doesnt exist, unique variable of 1st spectrum
            # if it fails, let it fail
    # Make sure it is in both Spectra
    if var not in s1.get_vars():
        raise KeyError('Variable {0} not in Spectrum {1}'.format(var, s1.get_name()))
    if var not in s2.get_vars():
        raise KeyError('Variable {0} not in Spectrum {1}'.format(var, s1.get_name()))
    
    # Use same units 
    Iunit = s1.units[var]            
    wunit = s1.get_waveunit()
    medium = s1.get_medium()

    # Resample
    if wunit == 'nm':
        s2 = s2.resample(s1.get_wavelength(), 'nm', inplace=False)
    elif wunit == 'cm-1':
        s2 = s2.resample(s1.get_wavenumber(), 'cm-1', inplace=False)
    else:
        raise ValueError('Unexpected wunit: {0}'.format(wunit))
    
    # Substract
    w1, I1 = s1.get(var=var, Iunit=Iunit, wunit=wunit, medium=medium)
    w2, I2 = s2.get(var=var, Iunit=Iunit, wunit=wunit, medium=medium)

    name = s1.get_name()+'-'+s2.get_name()
    
    sub = Spectrum.from_array(w1, I1 - I2, var, 
                               waveunit=wunit, 
                               unit=Iunit,
                               conditions={'medium' : medium}, 
                               name=name)
#    warn("Conditions of the left spectrum were copied in the substraction.", Warning)
    return sub


def concat_spectra(s1, s2, var=None):
    ''' Concatenate two spectra ``s1`` and ``s2`` side by side.
    
    Note: their spectral range should not overlap 
    
    Returns    
    -------
    
    s: Spectrum
        Spectrum object with the same units and waveunits as ``s1``
        
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
        Spectrum you want to concatenate
    var: str
        quantity to manipulate: 'radiance', 'transmittance', ... If ``None``, 
        get the unique spectral quantity of ``s1``, or the unique spectral
        quantity of ``s2``, or raises an error if there is any ambiguity
        
    Notes
    -----
    
    .. warning::
        
        the output Spectrum has the sum of the spectral ranges of s1 and s2. 
        It won't be evenly spaced. This means that you cannot apply a slit without
        side effects. Typically, you want to use this function for convolved 
        quantities only, such as experimental spectra. Else, use
        :func:`~radis.los.slabs.MergeSlabs` with the options 
        ``resample='full', out='transparent'``
        
    See Also
    --------
    
    :func:`~radis.spectrum.operations.add_spectra`, 
    :func:`~radis.los.slabs.MergeSlabs`
    
    '''
    
    # Get variable
    if var is None:
        try:
            var = _get_unique_var(s2, var, inplace=False)    # unique variable of 2nd spectrum  
        except KeyError:
            var = _get_unique_var(s1, var, inplace=False)    # if doesnt exist, unique variable of 1st spectrum
            # if it fails, let it fail
    # Make sure it is in both Spectra
    if var not in s1.get_vars():
        raise KeyError('Variable {0} not in Spectrum {1}'.format(var, s1.get_name()))
    if var not in s2.get_vars():
        raise KeyError('Variable {0} not in Spectrum {1}'.format(var, s1.get_name()))
        
    if var in ['transmittance_noslit', 'transmittance']:
        warn('It does not make much physical sense to sum transmittances. Are '+\
             'you sure of what you are doing? See also // (MergeSlabs) and > '+\
             '(SerialSlabs)')
    
    # Use same units 
    Iunit = s1.units[var]            
    wunit = s1.get_waveunit()
    medium = s1.get_medium()

    # Get the value, on the same wunit)
    w1, I1 = s1.get(var=var, copy=False)  # @dev: faster to just get the stored value. 
                                          # it's copied in hstack() below anyway). 
    w2, I2 = s2.get(var=var, Iunit=Iunit, wunit=wunit, medium=medium)
    
    if not (w1.max() < w2.min() or w2.max() > w1.min()):
        raise ValueError("You cannot use concat_spectra for overlapping spectral ranges. "+\
                         "Got: {0:.2f}-{1:.2f} and {2:.2f}-{3:.2f} {4}. ".format(w1.min(), w1.max(),
                                                               w2.min(), w2.max(),
                                                               wunit)+\
                         "Use MergeSlabs instead, with the correct `out=` parameter "+\
                         "for your case")
                         
    w_tot = hstack((w1, w2))
    I_tot = hstack((I1, I2))

    name = s1.get_name()+'&'+s2.get_name()     # use "&" instead of "+" 
    
    concat = Spectrum.from_array(w_tot, I_tot, var, 
                               waveunit=wunit, 
                               unit=Iunit,
                               conditions={'medium' : medium}, 
                               name=name)
    
    return concat
    

def offset(s, offset, unit, name=None, inplace=False):
    # type: (Spectrum, float, str, str, bool) -> Spectrum
    '''Offset the spectrum by a wavelength or wavenumber 

    Parameters    
    ----------
    s: Spectrum
        Spectrum you want to modify
    offset: float
        Constant to add to all quantities in the Spectrum.
    unit: 'nm' or 'cm-1'
        unit for ``offset``.
    name: str
        name of output spectrum
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy. 
        Default ``False``

    Returns    
    -------
    s : Spectrum
        Spectrum object where cst is added to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.
        
    See Also
    --------
    
    call as a Spectrum method directly: :py:meth:`~radis.spectrum.spectrum.Spectrum.offset`
    '''
    
    has_var = len(s._q)>0
    has_conv_var = len(s._q_conv)>0
    
    # Convert to correct unit:
    if unit == 'nm' and s.get_waveunit() == 'cm-1':
        # Note @EP: technically we should check the medium is air or vacuum...  TODO
        # Note @EP: here we're offsetting by a constant value in 'cm-1', which is
        # not a constant value in 'nm'. We end of with offset as an array 
        if has_var: offset_q = - dnm2dcm(offset, s.get_wavelength(which='non_convoluted'))  # this is an array
        if has_conv_var: offset_qconv = - dnm2dcm(offset, s.get_wavelength(which='convoluted'))  # this is an array
    elif unit == 'cm-1' and s.get_waveunit() == 'nm':
        if has_var: offset_q = - dcm2dnm(offset, s.get_wavenumber(which='non_convoluted'))  # this is an array
        if has_conv_var: offset_qconv = - dcm2dnm(offset, s.get_wavenumber(which='convoluted'))  # this is an array
    else:
        assert unit == s.get_waveunit()
        if has_var: offset_q = offset
        if has_conv_var: offset_qconv = offset
        
    if not inplace:
        s = s.copy()
        
    # Update all variables
    if has_var:
        s._q['wavespace'] += offset_q
        # @dev: updates the Spectrum directly because of copy=False
    if has_conv_var:
        s._q_conv['wavespace'] += offset_qconv
        
    if name:
        s.name = name
        
    return s

def get_baseline(s, var='radiance', wunit='nm', medium='air', Iunit=None):
    '''Calculate and returns a baseline 

    Parameters    
    ----------
    s: Spectrum
        Spectrum which needs a baseline

    Returns    
    -------
    baseline: Spectrum
        Spectrum object where intenisity is the baseline of s is computed by peakutils
        
    See Also
    --------
    
    :py:func:`~radis.spectrum.operations.sub_baseline`
    '''
    import peakutils 
    w1, I1 = s.get(var=var, Iunit=Iunit, wunit=wunit, medium=medium)
    baseline = peakutils.baseline(I1, deg=1, max_it = 500)
    baselineSpectrum = Spectrum.from_array(w1, baseline, var, 
                               waveunit=wunit, 
                               unit=Iunit,
                               conditions={'medium' : medium}, 
                               name=s.get_name()+'_baseline')
    return baselineSpectrum
# %% Tests

def test_multiplyAndAddition(*args, **kwargs):

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update('radiance_noslit', verbose=False)
    s.apply_slit(0.1)
    s = Radiance(s)
    assert s.units['radiance'] == 'mW/cm2/sr/nm'

    s_bis = add_constant(s, 1, 'mW/cm2/sr/nm')
    w, Idiff = get_diff(s_bis, s, 'radiance') 
    test = Idiff[1]-1
    assert np.all(test<1e-10)

    s_ter = multiply(multiply(s, 50), 1/50)
#    plot_diff(s_ter, s_5)
    diff = get_diff(s_ter, s, 'radiance') 
    ratio = abs(np.trapz(diff[1], x=diff[0])/s.get_integral('radiance'))
    assert ratio<1e-10

def test_visualTestBaseline(plot=True, *args, **kwargs):

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update('radiance_noslit', verbose=False)
    s.apply_slit(0.1)
    s = Radiance_noslit(s)
    assert s.units['radiance'] == 'mW/cm2/sr/nm'
    
    s2 = sub_baseline(s, 2e-4, -2e-4, name = 'sub_arb_baseline')
    if plot:
        plot_diff(s, s2)
    # TODO: add Test on intensity on both sides?

def test_offset(plot=True):
    
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update('radiance_noslit', verbose=False)
    s.apply_slit(0.1)
    
    s2 = offset(s, 10, 'nm', name = 'offset_10nm')
    if plot:
        plot_diff(s, s2)
    assert np.allclose(s2.get_wavelength(which='convoluted'), 
                       s.get_wavelength(which='convoluted')+10)
    assert np.allclose(s2.get_wavelength(which='non_convoluted'), 
                       s.get_wavelength(which='non_convoluted')+10)
    
    # Test inplace version
    s.offset(10, 'nm')
    assert np.allclose(s2.get_wavelength(which='convoluted'), 
                       s.get_wavelength(which='convoluted'))
    

if __name__ == '__main__':
    
    from radis import load_spec, get_diff, plot_diff
    from radis.test.utils import getTestFile
    import numpy as np
    import pytest
   
    
    test_multiplyAndAddition()
    test_offset()
    
    # An implement of Spectrum Algebra
    # Reload:
    s=load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    s.update()

    # Test addition of Spectra
    s.plot(lw=2, nfig='Merge: s//s')
    (s//s).plot(nfig='same')
    
    # Test substraction of Spectra
    s_tr = Transmittance_noslit(s)
    assert (s_tr-1.0*s_tr).get_integral('transmittance_noslit') == 0

    # TODO: add test
    # @EP: the test fails at the moment because multiply only works with radiance,
    # and MergeSlabs only works with non convoluted quantities
    # Do we want that? Up for discussion...
    
    # There should be an error if algebraic operations are used when 
    # multiple quantities are defined:
    with pytest.raises(KeyError):
        2*s
        
    s.apply_slit(0.1, 'nm')
    s_rad = Radiance(s)
    
    # Test multiplication with float
    s.plot(lw=2, nfig='Multiplication (by scalar): 2*s', wunit='nm')
#    (s*s).plot(nfig='same')
    (2*s_rad).plot(nfig='same', wunit='nm')
    
    # Test Serial:
    s.rescale_path_length(20)
    s.plot('radiance_noslit', lw=2, nfig='Line of sight (SerialSlabs): s > s > s')
    (s>s>s).plot('radiance_noslit', nfig='same')

    # Test algebraic addition (vs multiplication)
    assert (s_rad+s_rad).compare_with(2*s_rad, spectra_only='radiance', plot=False)

    # Test algebraic addition with different waveunits
    s_rad_nm = s_rad.resample(s_rad.get_wavelength(), 'nm', inplace=False)
    s_sum = (2*s_rad_nm-s_rad_nm)
    s_sum.compare_with(s_rad, spectra_only='radiance', plot=True)
    assert (s_rad_nm+s_rad_nm).compare_with(2*s_rad, spectra_only='radiance', plot=True,rtol=1e-3)
    


# -*- coding: utf-8 -*-
"""
Created on Tue Jan  9 17:14:01 2018

@author: erwan

Test broadening against HAPI and tabulated data

We're looking at CO(0->1) line 'R1' at 2150.86 cm-1

"""

from __future__ import unicode_literals, print_function, absolute_import, division
from radis.lbl.factory import SpectrumFactory
from radis.spectrum.spectrum import Spectrum
from radis import plot_diff, get_residual_integral, get_residual
from radis.misc.utils import DatabankNotFound
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
from radis.misc.printer import printm
from os.path import join, dirname
import matplotlib.pyplot as plt
import pytest


@pytest.mark.fast
def test_broadening(rtol=1e-2, verbose=True, plot=False, *args, **kwargs):
    '''
    Test broadening against HAPI and tabulated data

    We're looking at CO(0->1) line 'R1' at 2150.86 cm-1
    '''
    from radis.io.hapi import db_begin, fetch, tableList, absorptionCoefficient_Voigt

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

    # Conditions
    T = 3000
    p = 0.0001
    wstep = 0.001
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    broadening_max_width = 10  # cm-1

    # %% HITRAN calculation
    # -----------

    # Generate HAPI database locally
    db_begin(join(dirname(__file__), __file__.replace('.py', '_HAPIdata')))
    if not 'CO' in tableList():   # only if data not downloaded already
        fetch('CO', 5, 1, wmin-broadening_max_width /
              2, wmax+broadening_max_width/2)
        # HAPI doesnt correct for side effects

    # Calculate with HAPI
    nu, coef = absorptionCoefficient_Voigt(SourceTables='CO',
                                           Environment={'T': T,  # K
                                                        'p': p/1.01325,  # atm
                                                        },
                                           WavenumberStep=wstep,
                                           HITRAN_units=False)

    s_hapi = Spectrum.from_array(nu, coef, 'abscoeff', 'cm-1', 'cm_1',
                                 conditions={'Tgas': T},
                                 name='HAPI')

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,   # doesnt change anything
        wstep=wstep,
        pressure=p,
        broadening_max_width=broadening_max_width,
                         isotope=[1])  # 0.2)
    sf.load_databank('HITRAN-CO-TEST')
#    s = pl.non_eq_spectrum(Tvib=T, Trot=T, Ttrans=T)
    s = sf.eq_spectrum(Tgas=T, name='RADIS')

    # Plot and compare
    res = abs(get_residual_integral(s, s_hapi, 'abscoeff'))
    if plot:
        plot_diff(s, s_hapi, var='abscoeff',
                  title='{0} bar, {1} K (residual {2:.2g}%)'.format(p, T, res*100), show_points=False)
        plt.xlim((wmin, wmax))
    if verbose:
        printm('residual:', res)
    assert res < rtol


@pytest.mark.fast
def test_voigt_broadening_methods(verbose=True, plot=False, *args, **kwargs):
    '''
    Test direct Voigt broadening vs convolution of Gaussian x Lorentzian
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

    # Conditions
    T = 3000
    p = 1
    wstep = 0.1
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    broadening_max_width = 10  # cm-1
    
    for i, wstep in enumerate([0.01, 0.1, 0.5]):

        # %% Calculate with RADIS
        # ----------
        sf = SpectrumFactory(
            wavenum_min=wmin,
            wavenum_max=wmax,
            mole_fraction=1,
            path_length=1,   # doesnt change anything
            wstep=wstep,
            pressure=p,
            broadening_max_width=broadening_max_width,
            isotope='1',
            verbose=False,
            )  # 0.2)
        sf.load_databank('HITRAN-CO-TEST')
    #    s = pl.non_eq_spectrum(Tvib=T, Trot=T, Ttrans=T)
        sf._broadening_method = 'voigt'
        s_voigt = sf.eq_spectrum(Tgas=T, name='direct')
        
        sf._broadening_method = 'convolve'
        s_convolve = sf.eq_spectrum(Tgas=T, name='convolve')
    
        res = get_residual(s_voigt, s_convolve, 'abscoeff')
        
        if verbose:
            print('Residual:', res)

        # plot the last one    
        if plot:
            plot_diff(s_voigt, s_convolve, 'abscoeff', nfig='test_voigt_broadening_methods'+str(i),
                      title='P {0} bar, T {1} K, wstep {2} cm-1'.format(p, T, wstep))
        
        assert res < 2e-4
        

#@pytest.mark.needs_config_file
#@pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_abscoeff_continuum(plot=False, verbose=2, warnings=True, *args, **kwargs):
    ''' 
    Test calculation with pseudo-continuum

    Assert results on abscoeff dont change
    
    
    Notes
    -----
    
    Uses HITRAN so it can deployed and tested on [Travis]_, but we should switch
    to HITEMP if some HITEMP files can be downloaded automatically at the 
    execution time. 

    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:
        if verbose:
            printm('>>> test_abscoeff_continuum')

        sf = SpectrumFactory(wavelength_min=4200, wavelength_max=4500,
                             parallel=False, bplot=False, cutoff=1e-23,
                             molecule='CO2',
                             isotope='1,2', 
                             db_use_cached=True,
                             broadening_max_width=10,
                             path_length=0.1,
                             mole_fraction=1e-3,
                             medium='vacuum',
                             verbose=verbose)
        sf.warnings.update({
            'MissingSelfBroadeningWarning': 'ignore',
            'NegativeEnergiesWarning': 'ignore',
            'LinestrengthCutoffWarning': 'ignore'})
        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install
#        sf.load_databank('HITEMP-CO2-DUNHAM')       # to take a real advantage of abscoeff continuum, should calculate with HITEMP
        sf._export_continuum = True   # activate it 

        # Calculate one without pseudo-continuum
        sf.params.pseudo_continuum_threshold = 0
        s1 = sf.eq_spectrum(Tgas=2000)
        s1.name = 'All lines resolved ({0}) ({1:.1f}s)'.format(s1.conditions['lines_calculated'],
                                                               s1.conditions['calculation_time'])
        assert s1.conditions['pseudo_continuum_threshold'] == 0

        # Calculate one with pseudo-continuum
        sf.params.pseudo_continuum_threshold = 0.05
        s2 = sf.eq_spectrum(Tgas=2000)
        s2.name = 'Semi-continuum + {0} lines ({1:.1f}s)'.format(s2.conditions['lines_calculated'],
                                                                 s2.conditions['calculation_time'])
        assert s2.conditions['pseudo_continuum_threshold'] == 0.05
        assert 'abscoeff_continuum' in s2.get_vars()

        # Plot
        if plot:
            plot_diff(s1, s2, 'radiance_noslit', Iunit='µW/cm2/sr/nm',
                      nfig='test_abscoeff_continuum: diff')

            s2.plot('abscoeff', label=s2.name,
                    nfig='test_abscoeff_continuum: show continuum')
            s2.plot('abscoeff_continuum', nfig='same', label='Pseudo-continuum (aggreg. {0:g} lines)'.format(
                    s2.conditions['lines_in_continuum']),
                    force=True)

        # Compare
        res = get_residual(s1, s2, 'abscoeff')
        if verbose:
            printm('residual:', res)

        assert res < 1e-6

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)

#@pytest.mark.needs_config_file
#@pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_noneq_continuum(plot=False, verbose=2, warnings=True, *args, **kwargs):
    ''' 
    Test calculation with pseudo-continuum under nonequilibrium

    Assert results on emisscoeff dont change

    
    Notes
    -----
    
    Uses HITRAN so it can deployed and tested on [Travis]_, but we should switch
    to HITEMP if some HITEMP files can be downloaded automatically at the 
    execution time. 

    '''
   
    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:
        if verbose:
            printm('>>> test_noneq_continuum')

        sf = SpectrumFactory(wavelength_min=4200, wavelength_max=4500,
                             parallel=False, bplot=False, cutoff=1e-23,
                             molecule='CO2',
                             isotope='1,2', db_use_cached=True,
                             broadening_max_width=10,
                             path_length=0.1,
                             mole_fraction=1e-3,
                             medium='vacuum',
                             verbose=verbose)
        sf.warnings.update({
            'MissingSelfBroadeningWarning': 'ignore',
            'NegativeEnergiesWarning': 'ignore',
            'LinestrengthCutoffWarning': 'ignore'})
        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install
#        sf.load_databank('HITEMP-CO2-DUNHAM')       # to take a real advantage of abscoeff continuum, should calculate with HITEMP
        sf._export_continuum = True   # activate it 

        # Calculate one without pseudo-continuum
        sf.params.pseudo_continuum_threshold = 0
        s1 = sf.non_eq_spectrum(Tvib=2000, Trot=1000)
        s1.name = 'All lines resolved ({0}) ({1:.1f}s)'.format(s1.conditions['lines_calculated'],
                                                               s1.conditions['calculation_time'])
        assert s1.conditions['pseudo_continuum_threshold'] == 0

        # Calculate one with pseudo-continuum
        sf.params.pseudo_continuum_threshold = 0.05
        s2 = sf.non_eq_spectrum(Tvib=2000, Trot=1000)
        s2.name = 'Semi-continuum + {0} lines ({1:.1f}s)'.format(s2.conditions['lines_calculated'],
                                                                 s2.conditions['calculation_time'])
        assert s2.conditions['pseudo_continuum_threshold'] == 0.05
        assert 'abscoeff_continuum' in s2.get_vars()
        assert 'emisscoeff_continuum' in s2.get_vars()

        # Plot
        if plot:
            plot_diff(s1, s2, 'radiance_noslit', Iunit='µW/cm2/sr/nm',
                      nfig='test_noneq_continuum: diff')

            s2.plot('emisscoeff', label=s2.name,
                    nfig='test_noneq_continuum: show continuum')
            s2.plot('emisscoeff_continuum', nfig='same', label='Pseudo-continuum (aggreg. {0:g} lines)'.format(
                    s2.conditions['lines_in_continuum']), force=True)

        # Compare
        res = get_residual(s1, s2, 'abscoeff') + get_residual(s1, s2, 'emisscoeff')
        if verbose:
            printm('residual:', res)

        assert res < 5e-6

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)



def _run_testcases(plot=False, verbose=True, *args, **kwargs):

    # Test broadening
    test_broadening(plot=plot, verbose=verbose, *args, **kwargs)
    test_voigt_broadening_methods(plot=plot, verbose=verbose, *args, **kwargs)

    # Test pseudo-continuum
    test_abscoeff_continuum(plot=plot, verbose=verbose, *args, **kwargs)
    test_noneq_continuum(plot=plot, verbose=verbose, *args, **kwargs)

    return True


if __name__ == '__main__':
    printm('test_broadening: ', _run_testcases(
        plot=True, verbose=True, debug=False))
